//
// hf.cc
//
// Create a fractal landscape
//
// Copyright (c) J. Belson	1997.12.29
//
// $Author: jon $	:	$Date: 2003/07/26 17:01:33 $
//

#include <cassert>
#include <cstdio>
#include <cmath>

#include "hf.h"


/**
 * Constructor
 */
hf::hf(void)
{
	my_settings = settings::get_instance();

	// Seed rng
	seed = 1;

	offset = 0;

	// Slight kludge for now, size isn't relevent
	map_size = 256;
	map_order = 8;
}


/**
 * Destructor
 */
hf::~hf()
{
	;
}


/**
 * Scale noise according to terrain size
 */
float hf::scale(float f)
{
	return f * 0.20*TERRAIN_WIDTH;
}


/**
 * Apply offset to calculated heights
 */
void hf::set_offset(float f)
{
	offset = f;
}


/**
 * Interpolation routine, calls one of the routines below
 * @param a		First number
 * @param b		Second number
 * @param x		Interpolation quotiant, range 0.0 to 1.0
 */
float hf::interpolate(float a, float b, float x)
{
	return cosine_interpolate(a, b, x);
}


/**
 * Cosine interpolation
 */
float hf::cosine_interpolate(float a, float b, float x) 
{
	float ft = x * 3.1415927;
	float f = (1.0 - cos(ft)) * 0.5;

	return  a*(1.0-f) + b*f;
}


/**
 * Linear interpolation
 */
float hf::linear_interpolate(float a, float b, float x)
{
	return  a*(1.0-x) + b*x;
}


/**
 * Return approximation to surface normal at specified position
 */
vector3d hf::get_normal(float x, float y)
{
	static const float small = 0.1;

	float h1 = get_height(x, y);
	float h2 = get_height(x+small, y);
	float h3 = get_height(x, y+small);

	if (h1 < 0.0) h1 = 0.0;
	if (h2 < 0.0) h2 = 0.0;
	if (h3 < 0.0) h3 = 0.0;

	point3d p1(x,		h1,	y);
	point3d p2(x+small,	h2,	y);
	point3d p3(x,		h3,	y+small);


	vector3d v12 = p2 - p1;
	vector3d v13 = p3 - p1;

	v12.normalise();
	v13.normalise();

//	vector3d normal = v12.vector_product(v13);
	vector3d normal = v13.vector_product(v12);

	normal.normalise();

	return normal;
}


/**
 * Trace ray to see if it intersects terrain.
 * @param p		Starting point.
 * @param view	Direction of ray.
 * @return		True if path intesects, with p set to the point
 */
bool hf::trace_path(point3d& p, const vector3d view)
{
	bool got_hit = false;

	vector3d step = view;
	step.normalise();
	step = view * 2.0;
	step.normalise();

	// Slight gap to avoid colliding with ourself...
	p = p + step;

	// While we're still within heightfield area...
	const int min = -TERRAIN_WIDTH/2;
	const int max = TERRAIN_WIDTH/2;
	while (p.x >= min && p.x < max && p.z >= min && p.z < max && p.y < 50 /*&& p.y>-0.01*/) {

		//float height = get_height_fast(p.x, p.z);
		float height = get_height(p.x, p.z);

		// Have we hit anything?
		if (p.y > 0.0 && height >= p.y) {
			got_hit = true;
			break;
		}

		p = p + step;
	}

	return got_hit;
}


/**
 * Seed random number generators.  This allows a particular noise
 *  field to be reproduced.
 */
void hf::seed_rand(unsigned long int s)
{
	seed = s;
	srand((unsigned int) s);
	srandom((unsigned int) s);
}


