//
// Perlin noise generator
//
// Copyright (C) J. Belson 2003.02.08
//
// $Id: hf_ridged_perlin.cc,v 1.9 2003/10/11 18:11:00 jon Exp $
//

#include <math.h>
#include <time.h>

#define WANT_LOGGING

#include "hf_ridged_perlin.h"
#include "logging.h"
#include "noise.h"
#include "settings.h"


/**
 * Contructor
 */
hf_ridged_perlin::hf_ridged_perlin()
	: hf()
{
	// Get required fractal dimension
	my_settings->get_float(RIDGED_FD, &H);

	int val;
	my_settings->get_int(RIDGED_ITERATIONS, &val);
	octaves = val;

	my_settings->get_float(RIDGED_LACUNARITY, &lacunarity);

	my_settings->get_float(RIDGED_GAIN, &gain);
	gain = 0.3 + gain/40.0;

	my_settings->get_float(RIDGED_SCALE, &scaling);
	scaling = 0.1 + scaling/120.0;

	exponent_array = new double[MAX_NUM_OCTAVES];

	/* precompute and store spectral weights */
	/* compute weight for each frequency */
	for (int i=0; i<MAX_NUM_OCTAVES; i++) {
		exponent_array[i] = pow( lacunarity, -i*H );
	}

	srandom(seed);

	noise = new my_noise();
}


/**
 * Destructor
 */
hf_ridged_perlin::~hf_ridged_perlin()
{
	delete noise;
	delete [] exponent_array;
}


/**
 * Evaluate at specified point.
 * @param	x, y	Point at which to evaluate function.
 * @return			Computer value at point.
 */
float hf_ridged_perlin::get_height(float x, float y)
{
	if (seed == 0) {
		return -0.1 + offset;
	}
	return ridged_perlin(x, y) + offset;
	//return hetero(x, y);
}



float hf_ridged_perlin::ridged_perlin(double x, double y)
{
	x *= scaling;
	y *= scaling;

	double result;

	offs = 1;
	///H = 1.0;
	///gain = 2;
	///octaves = 5;

	point2d point;
	point.x = x;
	point.y = y;

	/* get first octave */
	signal = noise->noise2( point );

	/* get absolute value of signal (this creates the ridges) */
	if ( signal < 0.0 )
		signal = -signal;

	/* invert and translate (note that "offs" should be ~= 1.0) */
	signal = offs - signal;
	/* square the signal, to increase "sharpness" of ridges */
	signal *= signal;
	/* assign initial values */
	result = signal;
	weight = 1.0;

	for( int i=1; weight>0.001 && i<octaves; i++ ) {
		point.x *= lacunarity;
		point.y *= lacunarity;

		/* weight successive contributions by previous signal */
		weight = signal * gain;
		if ( weight > 1.0 ) weight = 1.0;
		if ( weight < 0.0 ) weight = 0.0;

		signal = noise->noise2( point );

		if ( signal < 0.0 )
			signal = -signal;
		signal = offs - signal;
		signal *= signal;
		/* weight the contribution */
		signal *= weight;
		result += signal * exponent_array[i];
	}

	result *= 15;

	return( result );
}




#if 0
/*
 * Heterogeneous procedural terrain fucntion: stats by altitude method.
 * Evaluated at "point"; returns value stored in "value".
 *
 * Parameters:
 * 	"H" is the fractal increment
 * 	"lacunarity" is the gap between successive frequencies
 * 	"octaves" is the number of frequencies in the fBm
 * 	"offs" raises the terrain from sea level
 */
float hf_ridged_perlin::hetero(double x, double y)
{
x /= 35.0;
y /= 35.0;

	Vector point;
	point.x = x;
	point.y = y;	//1.0;
	point.z = 0;	//1;	//y;


	float		value, increment, frequency, remainder;
	int			i;
	//static		first = TRUE;
	//static double   exponent_array[MAX_OCTAVES];

	/* first unscaled octave of function; later octaves are scaled */
	value = offs + Noise2( point );
	point.x *= lacunarity;
	point.y *= lacunarity;
	point.z *= lacunarity;

	/* inner loop of spectral construction, where the fractal is built */
	for (i=1; i<octaves; i++) {
		/* obtain displaced noise value */
		increment = Noise2( point ) + offs;

		/* scale amplitude appropriately for this frequency */
		increment *= exponent_array[i];

		/* scale increment by current altitude of fucntion */

		increment *= value;

		/* add increment to "value" */
		value += increment;
		point.x *= lacunarity;
		point.y *= lacunarity;
		point.z *= lacunarity;

	} /* for */

	/* take care of remainder in "octaves" */
	remainder = octaves - (int)octaves;
	if ( remainder ) {
		/* "i" and spatial freq. are preset in loop above */
		/* note that the main loop code is made shorter here */
		increment = (Noise2( point ) + offs) * exponent_array[i];
		value += remainder * increment * value;
	}
value *= 15;
	return( value );

} /* Hetero_Terrain() */
#endif


#if 0
/**
 * Override basic seed_rand
 */
void hf_ridged_perlin::seed_rand(unsigned long int s)
{
	seed = s;
	srand(s);
	srandom(s);
}
#endif	// 0




