//
// Perlin noise generator
//
// Copyright (C) J. Belson 2003.07.26
//
// $Id: noise.cc,v 1.4 2003/09/17 20:59:35 jon Exp $
//

#include <assert.h>
#include <cmath>
#include <cstdlib>

#define WANT_LOGGING

#include "logging.h"
#include "noise.h"



/**
 * Contructor.
 */
my_noise::my_noise()
{
	Log("my_noise::my_noise()");

	done_init = false;
}


/**
 * @return Random number in range -1.0 to 1.0
 */
float my_noise::unit_rand(void)
{
	float val = float (random())/(RAND_MAX/2) - 1;
	assert(val >= -1.0 && val <= 1.0);
	return val;
}


/**
 * Normalise a 2d vector.
 * @param	val	Pointer to vector coordinates.
 */
void my_noise::normalise2(float* val)
{
	float mag = std::sqrt(val[0]*val[0] + val[1]*val[1]);
	val[0] /= mag;
	val[1] /= mag;
}


/**
 * Perform bicubic interpolation.
 * @param	f	Value to be interpolated, range 0.0 to 1.0 inclusive.
 * @return		Interpolated function.
 */
float my_noise::interpolate(float f)
{
	return (3.0 - 2.0*f)*f*f;
}


/**
 * Evaluate 2d perlin function at a point
 * @param x, y		Coordinate of point.
 * @return			Evaluated noise function.
 */
float my_noise::noise2(float x, float y)
{
	if (!done_init) {
		done_init = true;
		init();
	}

	x /= 25.0;
	y /= 25.0;

	// Find bordering grid coordinates
	int x0 = int (x+0x1000)&(TABLE_SIZE - 1);
	int y0 = int (y+0x1000)&(TABLE_SIZE - 1);
	int x1 = x0 + 1;
	int y1 = y0 + 1;

	// Get gradient indices from permutation table
	int x0y0 = perms[(x0 + perms[(y0%TABLE_SIZE)])%TABLE_SIZE];
	int x0y1 = perms[(x0 + perms[(y1%TABLE_SIZE)])%TABLE_SIZE];
	int x1y0 = perms[(x1 + perms[(y0%TABLE_SIZE)])%TABLE_SIZE];
	int x1y1 = perms[(x1 + perms[(y1%TABLE_SIZE)])%TABLE_SIZE];

	// Offsets to x and y from unit square
	float xx = x + 0x1000;
	float dx0 = (int(x+0x1000)&(TABLE_SIZE-1)) - int (x);
	dx0 = xx - int(xx);
	float dx1 = dx0 - 1.0;

	float yy = y + 0x1000;
	float dy0 = (int (y+0x1000)&(TABLE_SIZE-1)) - int (y);
	dy0 = yy - int(yy);
	float dy1 = dy0 - 1.0;

	// Compute the dot products
	float dp00 = gradient2[x0y0][0]*dx0 + gradient2[x0y0][1]*dy0;
	float dp10 = gradient2[x1y0][0]*dx1 + gradient2[x1y0][1]*dy0;
	float dp01 = gradient2[x0y1][0]*dx0 + gradient2[x0y1][1]*dy1;
	float dp11 = gradient2[x1y1][0]*dx1 + gradient2[x1y1][1]*dy1;

	// Get weight in x and y directions
	float sx = interpolate(dx0);
	float sy = interpolate(dy0);

	float a = dp00 + sx*(dp10 - dp00);
	float b = dp01 + sx*(dp11 - dp01);

	return a + sy*(b - a);
}


/**
 * Populate look up tables to speed up calculations
 */
void my_noise::init(void)
{
	for (int i=0; i<TABLE_SIZE; i++) {
		perms[i] = i;

		gradient1[i] = unit_rand();

		for (int j=0; j<2; j++) {
			gradient2[i][j] = unit_rand();
		}
		normalise2(gradient2[i]);

		for (int j=0; j<3; j++) {
			gradient3[i][j] = unit_rand();
		}
		//normalise3(gradient3[i]);

	}

	// Jumble them around a bit
	for (int i=0; i<TABLE_SIZE; i++) {
		int idx = random()%TABLE_SIZE;
		int tmp = perms[i];
		perms[i] = perms[idx];
		perms[idx] = tmp;
	}
}



