//
// perlin.cc
//
// Perlin noise generator
//
// Copyright (C) J. Belson	2000.02.19
//

#include "assert.h"
#include <cmath>
#include <cstdlib>
#include <cstdio>

#include "perlin.h"


const float perlin::DEFAULT_PERSISTENCE	= 0.5;
const float perlin::DEFAULT_FREQ_SCALE	= 1.0;
const int perlin::DEFAULT_ITERATIONS	= 4;


///
/// Three prime number look-up tables
///
const long int perlin::primes[3][NUM_PRIMES] = {
	{
		#include "primes1.cc"
	},
	{
		#include "primes2.cc"
	},
	{
		#include "primes3.cc"
	}
};


/// Constructor
perlin::perlin()
	: persistence(DEFAULT_PERSISTENCE),
	  freq_scale(DEFAULT_FREQ_SCALE),
	  iterations(DEFAULT_ITERATIONS),
	  width(DEFAULT_WIDTH)
{
	;
}


/**
 * Actual noise function
 */
inline float perlin::noise(long int x, long int y)
{
	int n = x + y * 57;
	n = (n<<13) ^ n;
	return ( 1.0 - ( (n * (n * n * prime1 + prime2) + prime3) & 0x7fffffff) / 1073741824.0);
}



///
/// Cosine interpolation
///
float perlin::interpolate(float a, float b, float x) 
{
	float ft = x * 3.1415927;
	float f = (1.0 - cos(ft)) * 0.5;

	return  a*(1.0-f) + b*f;
}


///
/// Simple linear interpolation
///
float perlin::linear_interpolate(float a, float b, float x)
{
	return  a*(1.0-x) + b*x;
}


///
/// Filtered noise
///
float perlin::smoothed_noise(int x, int y)
{
	float corners = ( noise(x-1, y-1)+noise(x+1, y-1)+noise(x-1, y+1)+noise(x+1, y+1) ) / 16.0;
	float sides   = ( noise(x-1, y)  +noise(x+1, y)  +noise(x, y-1)  +noise(x, y+1) ) /  8.0;
	float center  =  noise(x, y) / 4.0;
	return corners + sides + center;
}


///
/// Interpolate value at (float, float) from (int, int)
///
float perlin::interpolated_noise(float x, float y)
{
	int integer_X    = int(x);
	float fractional_X = x - integer_X;

	int integer_Y    = int(y);
	float fractional_Y = y - integer_Y;

	float v1 = smoothed_noise(integer_X,     integer_Y);
	float v2 = smoothed_noise(integer_X + 1, integer_Y);
	float v3 = smoothed_noise(integer_X,     integer_Y + 1);
	float v4 = smoothed_noise(integer_X + 1, integer_Y + 1);

	float i1 = interpolate(v1, v2, fractional_X);
	float i2 = interpolate(v3, v4, fractional_X);

	return interpolate(i1 , i2 , fractional_Y);
}



///
/// Return height for point x, y
///
/// Frequency should be compensated for map_size
///
float perlin::perlin_noise_2d(float x, float y) 
{
	float total = 0.0;
	float p = persistence;

	for (int i=0; i<iterations; i++) {

		float frequency = freq_scale* pow(2.0, i);
		float amplitude = pow(p, i);

//		frequency *= 256.0/width;
//		frequency /= 2.0;

		int set = sets[i];
		prime1 = primes[0][set];
		prime2 = primes[1][set];
		prime3 = primes[2][set];

		total += interpolated_noise(x * frequency, y * frequency) * amplitude;
	}

	return total;
}



///
/// Noise configuration functions
///
void perlin::set_iterations(int val)
{
	iterations = val;
}

/// Noise configuration functions
void perlin::set_persistence(float val)
{
	persistence = val;
}

/// Noise configuration functions
void perlin::set_freq_scale(float val)
{
	freq_scale = val;
}

/// Noise configuration functions
void perlin::set_width(int val)
{
	width = val;
}

///
/// Randomly select which primes we're using for
///  each iteration.
///
void perlin::set_seed(int val)
{
printf("perlin::set_seed(%d)\n", val);
	srand(val);
	for (int i=0; i<MAX_ITERATIONS; i++) {
		sets[i] = rand()%NUM_PRIMES;
	}
}


