//
// perlin2
//
// From 'Procedural Texturing and Modelling'
//
// My bits Copyright (C) J. Belson 2001.06.11
//


#include <cstdlib>
#include <cstdio>

#include "perlin2.h"

#define setup(i, b0, b1, r0, r1)	\
		t = vec[i] + 0x1000;		\
		b0 = (int (t)) & (B - 1);	\
		b1 = (b0 + 1) & (B - 1);	\
		r0 = t - int (t);			\
		r1 = r0 - 1;

#define s_curve(t) ( t * t * (3. - 2. * t) )

#define lerp(t, a, b) ( a + t * (b - a) )


const float perlin2::DEFAULT_FREQ_SCALE	= 1.0;


/// Normalise in two dimensions
static void normalize2(float v[2])
{
	float s;

	s = sqrt(v[0] * v[0] + v[1] * v[1]);
	v[0] = v[0] / s;
	v[1] = v[1] / s;
}

/// Normalise in three dimensions
static void normalize3(float v[3])
{
	float s;

	s = sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
	v[0] = v[0] / s;
	v[1] = v[1] / s;
	v[2] = v[2] / s;
}


///
/// Constructor
///
perlin2::perlin2()
	: freq_scale(DEFAULT_FREQ_SCALE),
	  done_init(false)
{
	;
}

///
/// Speed up generation by pre-calculating data
///
void perlin2::init(void)
{
	done_init = true;

	int i, j, k;

	for (i = 0 ; i < B ; i++) {
		p[i] = i;

		g1[i] = (float)((random() % (B + B)) - B) / B;

		for (j = 0 ; j < 2 ; j++)
			g2[i][j] = (float)((random() % (B + B)) - B) / B;
		normalize2(g2[i]);

		for (j = 0 ; j < 3 ; j++)
			g3[i][j] = (float)((random() % (B + B)) - B) / B;
		normalize3(g3[i]);
	}

	while (--i) {
		k = p[i];
		p[i] = p[j = random() % B];
		p[j] = k;
	}

	for (i = 0 ; i < B + 2 ; i++) {
		p[B + i] = p[i];
		g1[B + i] = g1[i];
		for (j = 0 ; j < 2 ; j++)
			g2[B + i][j] = g2[i][j];
		for (j = 0 ; j < 3 ; j++)
			g3[B + i][j] = g3[i][j];
	}
}


///
/// Calculate dot product
///
float perlin2::DOT(float* a, float* b)
{
	float f = a[0]*b[0] + a[1]*b[1] + a[2]*b[2];

	return f;
}



///
/// Calculate two dimensional noise at point given by vec[]
///
float perlin2::noise2(float vec[2])
{
	if (!done_init) {
		init();
	}

	vec[0] *= freq_scale;
	vec[1] *= freq_scale;

	int bx0, bx1, by0, by1, b00, b10, b01, b11;
	float rx0, rx1, ry0, ry1, *q, sx, sy, a, b, t, u, v;
	register int i, j;


	setup(0, bx0,bx1, rx0,rx1);
	setup(1, by0,by1, ry0,ry1);

	i = p[ bx0 ];
	j = p[ bx1 ];

	b00 = p[ i + by0 ];
	b10 = p[ j + by0 ];
	b01 = p[ i + by1 ];
	b11 = p[ j + by1 ];

	sx = s_curve(rx0);
	sy = s_curve(ry0);

#define at2(rx,ry) ( rx * q[0] + ry * q[1] )

	q = g2[ b00 ] ; u = at2(rx0,ry0);
	q = g2[ b10 ] ; v = at2(rx1,ry0);
	a = lerp(sx, u, v);

	q = g2[ b01 ] ; u = at2(rx0,ry1);
	q = g2[ b11 ] ; v = at2(rx1,ry1);
	b = lerp(sx, u, v);

	return lerp(sy, a, b);
}



///
/// Calculate three dimensional noise at point given by vec[]
///
float perlin2::noise3(float vec[3])
{
	if (!done_init) {
		init();
	}

	int bx0, bx1, by0, by1, bz0, bz1, b00, b10, b01, b11;
	float rx0, rx1, ry0, ry1, rz0, rz1, *q, sx, sy, sz, a, b, c, d, t, u, v;
	int i, j;
	
	setup(0, bx0, bx1, rx0, rx1);
	setup(1, by0, by1, ry0, ry1);
	setup(2, bz0, bz1, rz0, rz1);

	i = p[bx0];
	j = p[bx1];
	
	b00 = p[i + by0];
	b10 = p[j + by0];
	b01 = p[i + by1];
	b11 = p[j + by1];

#define at(rx, ry, rz) (rx*q[0] + ry*q[1] + rz*q[2])

	sx = s_curve(rx0);
	sy = s_curve(ry0);
	sz = s_curve(rz0);

	q = g3[b00 + bz0];	u = at(rx0, ry0, rz0);
	q = g3[b10 + bz0];	v = at(rx1, ry0, rz0);
	a = lerp(sx, u, v);
	
	q = g3[b01 + bz0];	u = at(rx0, ry1, rz0);
	q = g3[b11 + bz0];	v = at(rx1, ry1, rz0);
	b = lerp(sx, u, v);

	c = lerp(sy, a, b);			// Interpolate in y at low x
	q = g3[b00 + bz1];	u = at(rx0, ry0, rz1);
	q = g3[b10 + bz1];	v = at(rx1, ry0, rz1);
	a = lerp(sx, u, v);

	q = g3[b01 + bz1];	u = at(rx0, ry1, rz1);
	q = g3[b11 + bz1];	v = at(rx1, ry1, rz1);
	b = lerp(sx, u, v);

	d = lerp(sy, a, b);			// Interpolate in y at high x

	return 1.5*lerp(sz, c, d);	// Interpolate in z
}


void perlin2::set_seed(int val)
{
	srandom(val);
}


void perlin2::set_freq_scale(float val)
{
	freq_scale = val;
}







