//
// skylight.h
//
// Model for skylight based on sun position and turbidity
//
// Implemented from paper "A Practical Analytic Model for Daylight"
// A. J. Preetham, Peter Shirley, Brian Smits, Siggraph 99
// (I left out all the difficult bits :-)
//
// Copyright (C) J. Belson	2000.05.19
//
// $Author: jon $	:	$Date: 2003/05/30 11:04:04 $
//

#include <cmath>

#include "fcolour.h"
#include "skylight.h"
#include "vector3d.h"



skylight::skylight(double t)
{
	set_turbidity(t);

	lum_scale = 25;

	largest = 0;

	sun = vector3d(0.0, 1.0, 0.0);
	sun.normalise();
}



/**
 * Set direction to sun
 */
void skylight::set_sun_direction(const vector3d& v)
{
	std::cout << "Sun direction is " << v << std::endl;
	sun = v;
}



/**
 * Set atmospheric haze to be used for model
 */
void skylight::set_turbidity(float t)
{
	turbidity = t;

	// Precompute Perez coefficients for new turbidity
	c_lum	= coeff_luminance();
	c_x		= coeff_x();
	c_y		= coeff_y();
}



/**
 * Calculate colour for specified viewing direction
 */
fcolour skylight::get_colour(const vector3d& v)
{
	vector3d n(0.0, 1.0, 0.0);	// Normal to ground plane

	// 'theta' is angle between [0,1,0] and view vector
	double theta = n.angle(v);

	// 'theta_s' is angle between [0,1,0] and sun vector
	double theta_s = n.angle(sun);

	// 'gamma' is angle between view vector and sun vector
	double gamma = v.angle(sun);


	double zen_lum = zenith_luminance(theta_s);
	double lum = zen_lum * perez_function(theta, gamma, c_lum) / perez_function(0.0, theta_s, c_lum);

	double zen_x = zenith_x(theta_s);
	double x = zen_x * perez_function(theta, gamma, c_x) / perez_function(0.0, theta_s, c_x);

	double zen_y = zenith_y(theta_s);
	double y = zen_y * perez_function(theta, gamma, c_y) / perez_function(0.0, theta_s, c_y);

	static double largest = 0;
	if (lum > largest) {
		largest = lum;
		//printf("Largest is %f\n", lum/lum_scale);
	}

	float adj = lum/lum_scale;

	return cie_to_rgb(x, y, adj);
}



/**
 * Luminance coefficients for Perez model
 */
skylight::coeff skylight::coeff_luminance(void)
{
	coeff c;

	c.A =  0.1787 * turbidity - 1.4630;
	c.B = -0.3554 * turbidity + 0.4275;
	c.C = -0.0227 * turbidity + 5.3251;
	c.D =  0.1206 * turbidity - 2.5771;
	c.E = -0.0670 * turbidity + 0.3703;

	return c;
}



/**
 * x coefficients for Perez model
 */
skylight::coeff skylight::coeff_x(void)
{
	coeff c;

	c.A = -0.0193 * turbidity - 0.2592;
	c.B = -0.0665 * turbidity + 0.0008;
	c.C = -0.0004 * turbidity + 0.2125;
	c.D = -0.0641 * turbidity - 0.8989;
	c.E = -0.0033 * turbidity + 0.0452;

	return c;
}



/**
 * y coefficients for Perez model
 */
skylight::coeff skylight::coeff_y(void)
{
	coeff c;

	c.A = -0.0167 * turbidity - 0.2608;
	c.B = -0.0950 * turbidity + 0.0092;
	c.C = -0.0079 * turbidity + 0.2102;
	c.D = -0.0441 * turbidity - 1.6537;
	c.E = -0.0109 * turbidity + 0.0529;

	return c;
}



/**
 * Calculate zenith value of luminance for given sun position
 * @param theta sun azimuth, 0 = directly overhead
 */
double skylight::zenith_luminance(double theta)
{
	double val = (4.0/9.0 - turbidity/120.0)*(M_PI - 2.0*theta);

	double Yz = (4.0453*turbidity - 4.9710)*std::tan(val)
					- 0.2155*turbidity + 2.4192;

	return Yz;
}



/**
 * Calculate zenith value of 'x' for given sun position
 * @param theta sun azimuth, 0 = directly overhead
 */
double skylight::zenith_x(double theta)
{
	double a =	 0.00166 * turbidity*turbidity +
				-0.02903 * turbidity +
			 	 0.11693 * 1.0;

	double b =	-0.00375 * turbidity*turbidity +
				 0.06377 * turbidity +
				-0.21196 * 1.0;

	double c =	 0.00209 * turbidity*turbidity +
				-0.03202 * turbidity +
			 	 0.06052 * 1.0;

	double d =	 0.0     * turbidity*turbidity +
			  	 0.00394 * turbidity +
			  	 0.25886 * 1.0;


	return (a * theta*theta*theta	+
			b * theta*theta			+
			c * theta				+
			d * 1.0);
}



/**
 * Calculate zenith value of '' for given sun position
 * @param theta sun azimuth, 0 = directly overhead
 */
double skylight::zenith_y(double theta)
{
	double a =	 0.00275 * turbidity*turbidity +
				-0.04214 * turbidity +
				 0.15346 * 1.0;

	double b =	-0.00610 * turbidity*turbidity +
				 0.08970 * turbidity +
				-0.26756 * 1.0;

	double c =	 0.00317 * turbidity*turbidity +
				-0.04153 * turbidity +
				 0.06670 * 1.0;

	double d =	 0.00000 * turbidity*turbidity +
			 	 0.00516 * turbidity +
				 0.26688 * 1.0;


	return (a * theta*theta*theta	+
			b * theta*theta			+
			c * theta				+
			d * 1.0);
}



/**
 * Calculate the RGB values that correspond to our x, y and Y (luminance)
 */
fcolour skylight::cie_to_rgb(double x, double y, double lum) const
{
	// First calculate X, Y and Z
	double X = (x/y) * lum;
	double Y = lum;
	double Z = (lum/y)*(1.0 - x - y);

	fcolour col;

	col.r =  3.240479*X - 1.537150*Y - 0.498535*Z;
	col.g = -0.969256*X + 1.875991*Y + 0.041556*Z;
	col.b =  0.055648*X - 0.204043*Y + 1.057311*Z;


	// Deal with colours undisplayable using RGB model
	col.clamp();

	return col;
}



/**
 * Evaluate Perez function for given angles and turbidity
 */
double skylight::perez_function(double theta, double gamma, coeff c)
{
	double a = (1.0 + c.A*exp(c.B/std::cos(theta)));
	double b = (1.0 + c.C*exp(c.D*gamma) + c.E*std::cos(gamma)*std::cos(gamma));

	return a*b;
}



////////////////////////////////////////////////////////////////////////////////
// Some unused code to deal with RGB -> CIE colour conversion in a slightly
//  smarter way.  Unfortunately I don't have a copyright notice for this, but
//  I /think/ I got it from Tom Bech.


#if 0

static void xyz_to_rgb(struct colourSystem *cs, double xc, double yc, double zc,
					double *r, double *g, double *b);
static int constrain_rgb(struct colourSystem *cs, double *x, double *y, double *z,
				double *r, double *g, double *b);
static int inside_gamut(double r, double g, double b);


#define CLAMP(v, l, h)  ((v) < (l) ? (l) : (v) > (h) ? (h) : (v))
#define TRUE    1
#define FALSE   0

#define Maxval  255                   /* Maxval to use in generated pixmaps */

/* A  colour  system is defined by the CIE x and y  coordinates of its
   three primary illuminants and the x and y coordinates of the  white
   point. */

struct colourSystem {
    char *name;                       /* Colour system name */
    double xRed, yRed,                /* Red primary illuminant */
           xGreen, yGreen,            /* Green primary illuminant */
           xBlue, yBlue,              /* Blue primary illuminant */
           xWhite, yWhite;            /* White point */
};

static int interpwp = FALSE;          /* Interpolate to white point ? */

/* Standard white point chromaticities. */

#define IlluminantC     0.3101, 0.3162  /* For NTSC television */
#define IlluminantD65   0.3127, 0.3291  /* For EBU and SMPTE */

static struct colourSystem
    NTSCsystem  =  { "NTSC",               0.67,  0.33,  0.21,  0.71,  0.14,  0.08,  IlluminantC   },
    EBUsystem   =  { "EBU (PAL/SECAM)",    0.64,  0.33,  0.29,  0.60,  0.15,  0.06,  IlluminantD65 },
    SMPTEsystem =  { "SMPTE",             0.630, 0.340, 0.310, 0.595, 0.155, 0.070, IlluminantD65 },
    HDTVsystem  =  { "HDTV",              0.670, 0.330, 0.210, 0.710, 0.150, 0.060, IlluminantD65 },
    CIEsystem   =  { "CIE",              0.7355,0.2645,0.2658,0.7243,0.1669,0.0085, 0.3894,0.3324},
    Customsystem = { "Custom",             0.64,  0.33,  0.29,  0.60,  0.15,  0.06,  IlluminantD65 };

static struct colourSystem *cs = &EBUsystem;

/*                            XYZ_TO_RGB

    Given  an additive tricolour system CS, defined by the CIE x and y
    chromaticities of its three primaries (z is derived  trivially  as
    1-(x+y)),  and  a  desired chromaticity (XC, YC, ZC) in CIE space,
    determine the contribution of each primary in a linear combination
    which   sums  to  the  desired  chromaticity.   If  the  requested
    chromaticity falls outside the  Maxwell  triangle  (colour  gamut)
    formed  by the three primaries, one of the r, g, or b weights will
    be negative.  Use inside_gamut() to test for a  valid  colour  and
    constrain_rgb()  to  desaturate  an  outside-gamut  colour  to the
    closest representation within the available gamut. */

static void xyz_to_rgb(struct colourSystem *cs, double xc, double yc, double zc,
					double *r, double *g, double *b)
{
    double xr, yr, zr, xg, yg, zg, xb, yb, zb;

    xr = cs->xRed;    yr = cs->yRed;    zr = 1 - (xr + yr);
    xg = cs->xGreen;  yg = cs->yGreen;  zg = 1 - (xg + yg);
    xb = cs->xBlue;   yb = cs->yBlue;   zb = 1 - (xb + yb);

    *r = (-xg*yc*zb + xc*yg*zb + xg*yb*zc - xb*yg*zc - xc*yb*zg + xb*yc*zg) /
         (xr*yg*zb - xg*yr*zb - xr*yb*zg + xb*yr*zg + xg*yb*zr - xb*yg*zr);

    *g = (xr*yc*zb - xc*yr*zb - xr*yb*zc + xb*yr*zc + xc*yb*zr - xb*yc*zr) /
         (xr*yg*zb - xg*yr*zb - xr*yb*zg + xb*yr*zg + xg*yb*zr - xb*yg*zr);

    *b = (xr*yg*zc - xg*yr*zc - xr*yc*zg + xc*yr*zg + xg*yc*zr - xc*yg*zr) /
         (xr*yg*zb - xg*yr*zb - xr*yb*zg + xb*yr*zg + xg*yb*zr - xb*yg*zr);
}

/*                           INSIDE_GAMUT

    Test whether a requested colour is  within  the  gamut  achievable
    with  the  primaries  of  the current colour system.  This amounts
    simply  to  testing  whether   all   the   primary   weights   are
    non-negative. */

static int inside_gamut(double r, double g, double b)
{
    return r >= 0 && g >= 0 && b >= 0;
}

/*                          CONSTRAIN-RGB

    If  the  requested RGB shade contains a negative weight for one of
    the primaries, it lies outside the colour  gamut  accessible  from
    the  given  triple of primaries.  Desaturate it by mixing with the
    white point of the colour system so as to reduce the primary  with
    the  negative  weight  to zero.  This is equivalent to finding the
    intersection on the CIE diagram of a line drawn between the  white
    point  and  and  the  requested  color and the edge of the Maxwell
    triangle formed by the three primaries.  If INTERPWP  is  nonzero,
    the   white  point  defined  by  the  sum  of  the  three  primary
    illuminants is used instead of the colour  system's  actual  white
    point.   While  indefensible from the standpoint of colour theory,
    this produces much better looking  charts  on  computer  monitors,
    since the white points of most colour systems are well to the blue
    of the white defined by the R = G = B = 1. */

static int constrain_rgb(struct colourSystem *cs, double *x, double *y, double *z,
				double *r, double *g, double *b)
{
    /* Is the contribution of one of the primaries negative ? */

    if (!inside_gamut(*r, *g, *b)) {
        double par, wr, wg, wb, xw, yw;

        /* Determine  the  white  point  used in interpolating  out of
           gamut colours.  If INTERPWP is  set,  the  colour  system's
           white  point  is  used.  Otherwise, the white defined by an
           equal mix of the three illuminants is taken as  the  origin
           of the interpolation line drawn to the out of gamut colour. */

        if (interpwp) {
            xw = cs->xWhite;
            yw = cs->yWhite;
        } else {
            xw = (cs->xRed + cs->xGreen + cs->xBlue) / 3;
            yw = (cs->yRed + cs->yGreen + cs->yBlue) / 3;
        }

        /* Yes.  Find the RGB mixing weights of the white point (we
           assume the white point is in the gamut!). */

        xyz_to_rgb(cs, xw, yw, 1 - (xw + yw), &wr, &wg, &wb);

        /* Find the primary with negative  weight  and  calculate  the
           parameter  of  the point on the vector from the white point
           to the original requested colour in RGB space. */

        if (*r < *g && *r < *b) {
            par = wr / (wr - *r);
        } else if (*g < *r && *g < *b) {
            par = wg / (wg - *g);
        } else {
            par = wb / (wb - *b);
        }

        /* Since XYZ space is a linear transformation of RGB space, we
           can  find  the XYZ space coordinates of the point where the
           edge of the gamut intersects  the  vector  from  the  white
           point  to  the original colour by multiplying the parameter
           in RGB space by the difference vector in XYZ space. */

        *x = CLAMP(xw + par * (*x - xw), 0, 1);
        *y = CLAMP(yw + par * (*y - yw), 0, 1);
        *z = CLAMP(1 - (*x + *y), 0, 1);

        /* Now finally calculate the gamut-constrained RGB weights. */

        *r = CLAMP(wr + par * (*r - wr), 0, 1);
        *g = CLAMP(wg + par * (*g - wg), 0, 1);
        *b = CLAMP(wb + par * (*b - wb), 0, 1);
        return 1;                     /* Colour modified to fit RGB gamut */
    }
    return 0;                         /* Colour within RGB gamut */
}




#endif	// 0









