//
// base_GLwindow.cc
//
// Window that supports GL rendering
//
// Copyright (c) J. Belson	2000.11.04
//
// $Author: jon $	:	$Date: 2003/10/11 18:11:00 $
//

#include <ctime>

#include "base_GLwindow.h"
#include "colour.h"
#include "hfield.h"
#include "hf_skydome.h"
#include "matrix.h"
#include "point3d.h"
#include "procedural.h"
#include "settings.h"
#include "stats.h"
#include "texture_cloud.h"
#include "texture_rough.h"
#include "texture_skydome.h"
#include "texture_skylight.h"
#include "types.h"
#include "util.h"

using std::cout;
using std::endl;

const GLfloat base_GLwindow::fog_colour[] = { 0.75, 0.75, 0.75, 1.0 };
const GLfloat base_GLwindow::sky_colour[] = { 82/255.0, 134/255.0, 214/255.0, 1.0 };


/**
 * Constructor
 */
base_GLwindow::base_GLwindow(int w, int h)
	: width(w),
	  height(h)
{
	my_settings = settings::get_instance();

	pos_x = pos_y = pos_z = 0.0;
	rot_a = rot_b = rot_c = 0.0;

	set_detail(8);

	// Set up viewing frustrum
	fov = 60.0/360 * 2*M_PI;
	view_frustrum = frustrum(fov, 4, 640);

	texturing = true;
}


/**
 * Destructor
 */
base_GLwindow::~base_GLwindow()
{
	;
}


/**
 * Set up OpenGL for rendering
 */
void base_GLwindow::GLinit(void)
{
	puts("GLinit()");

	// Start with clean slate...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// Take your pick of projections
//	glOrtho(-100, 100, -40, 40, -40, 80);
//	glOrtho(-100, 100, -40, 40, 40, -80);
//	glFrustum(-30,30, -30,30, 10,50);

	/// @todo fov set to 40 - does this match render view?  Should be a
	///  global setting.
	gluPerspective(40, 4.0/3.0, 4, /*320*/ 640);


	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();


	// Fix viewport to window size
	glViewport(0, 0, width, height);


	// Ambient light for scene
	GLfloat light_model_ambient[]	= { 0.25, 0.25, 0.25, 1.0 };
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, light_model_ambient);

	// Sun
	float ambient;
	my_settings->get_float(AMBIENT_INTENSITY, &ambient);

//	GLfloat light_position[]		= { light_i, light_j, light_k, 0.0 };
//	GLfloat light_diffuse[]			= { 0.75, 0.75,  0.75 };
//	GLfloat light_ambient[]			= { ambient, ambient, ambient  };

//	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
//	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
//	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);


		glEnable(GL_LIGHTING);
		glEnable(GL_LIGHT0);



	// Fog...
	glFogi(GL_FOG_MODE, GL_LINEAR /*GL_EXP*/);
	glFogfv(GL_FOG_COLOR, fog_colour);
	glFogf(GL_FOG_DENSITY, /*0.35*/ 0.01);
	glHint(GL_FOG_HINT, GL_DONT_CARE);
	glFogf(GL_FOG_START, 0.0 /*15.0*/);
	glFogf(GL_FOG_END, 400.0);

	glDisable(GL_FOG);

	// enable backface culling
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);


	glEnable(GL_DEPTH_TEST);

	// Default to smooth shading
	glShadeModel(GL_SMOOTH);

	// Vaguely sky cloured background
	glClearColor(sky_colour[0], sky_colour[1], sky_colour[2], 1.0);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


	gluLookAt(	0, 0.0,	0,		// Camera here
				0, 0, -1,		// looking here
				0, 1, 0);		// this way up

}


/**
 * Set viewpoint and viewing direction
 * @param eye		Eye position
 * @param target	Point to look at
 */
void base_GLwindow::set_view(const point3d& eye, const point3d& target)
{
	viewpoint = eye;

	// Set new view
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	gluLookAt(	eye.x, eye.y, eye.z,
				target.x, target.y, target.z,
				0, 1, 0);

	// Get currert model view matrix so we can use it for culling
	float elem[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, elem) ;
	camera = matrix(elem);
}


/**
 * Create array of points and surface normals for the sky
 */
void base_GLwindow::create_sky(const hfield* hf)
{
	// Get current sky prefs
	my_settings->get_colour(SKY_HORIZON_COLOUR, &horizon_colour);
	my_settings->get_colour(SKY_ZENITH_COLOUR, &zenith_colour);

	float i, j, k;
	my_settings->get_float3(LIGHT_RAY_VECTOR, &i, &j, &k);
	vector3d light = vector3d(i, j, k);

	bool realistic_sky;
	my_settings->get_bool(SKY_REALISTIC, &realistic_sky);

	sky.clear();

	sky.points		= new point3d[map_width*map_width];
	sky.normals		= new vector3d[map_width*map_width];

	// Convert heightfield into array of point3d.
	int terrain_width	= hf->get_width();
	const int min		= -(terrain_width)/2;
//	const int max		= (terrain_width)/2;
	const float step	= float(terrain_width)/(map_width-1);
	int dest			= 0;
	int src				= 0;

	fcolour col;

	textset_cloud ts;
	textset_cloud::load_settings(ts);
	texture* cloud = new texture_cloud(ts);

	texture* skydome;

	if (realistic_sky) {
		textset_skylight ts;
		ts.sun_direction = -light;
		ts.sun_direction.normalise();
		skydome = new texture_skylight(ts);
	} else {
		textset ts;
		ts.bg = zenith_colour;
		ts.fg = horizon_colour;
		skydome = new texture_skydome(ts);
	}

	float z = min;
	for (int j=0; j<map_width; j++) {
		float x = min;
		for (int i=0; i<map_width; i++) {

			point3d p(x, 0, z);
			float height = hf->get_height(p.x, p.z);

			p.y = height;

			col = skydome->colour(x, z, height);
			col.alpha = 1.0;
			fcolour col2 = cloud->colour(x, z, height);
			col = col.blend(col2);

			p.set_colour(col);

			sky.points[dest] = p;
			sky.normals[dest] = vector3d(0, 1, 0);

			src++;
			dest++;

			x += step;
		}
		z += step;
	}

	delete cloud;
	delete skydome;
}


/**
 * Create array of points and surface normals for the land
 */
void base_GLwindow::create_land(const hfield* hf)
{
	land.clear();

	land.points		= new point3d[map_width*map_width];
	land.normals	= new vector3d[map_width*map_width];

	// Convert heightfield into array of point3d.
	int terrain_width	= hf->get_width();
	const int min		= -(terrain_width)/2;
	//const int max		=  (terrain_width)/2;
	const float step	= float(terrain_width)/(map_width /*-1*/);
	int dest			= 0;
	int src				= 0;

	float water_level;
	my_settings->get_float(WATER_LEVEL, &water_level);

	// Prepare for calculating colours for vertices
	colour::init_colour();

	fcolour col;

	textset_noise ts;
	ts.seed = std::clock();
	texture* text = new texture_rough(ts);

	float z = min;
	for (int j=0; j<map_width; j++) {
		float x = min;
		for (int i=0; i<map_width; i++) {

			vector3d norm;

			point3d p(x, 0, z);
			float height = hf->get_height(p.x, p.z);

			// Flatten at sea level
			if (height < -0.1) {
				height = -0.15;
				col = colour::get_colour_from_height(height);
			} else {
				if (texturing) {
					col = text->colour(x*800, z*800, 0, &norm);
				} else {
					col = colour::get_colour_from_height(height);
				}
			}

			p.y = height;

			p.set_colour(col);

			land.points[dest] = p;
			land.normals[dest] = norm;

			src++;
			dest++;

			x += step;
		}
		z += step;
	}

	delete text;

	// This gets around the increasing lightness with increasing polygons problem
	// Unfortunately it overwrites any normal perturbations caused by the texture
	util::get_normal(map_width, land.points, land.normals);

	cout <<" Regenerated land, " << dest << " triangles" << endl;
}


/**
 * Render land and sky as requested
 */
void base_GLwindow::render(bool draw_land, bool draw_sky)
{
	cout << "GLview::render" << endl;

	assert(land.points);
	assert(land.normals);


	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glPushMatrix();

	fcolour colour;
	my_settings->get_colour(AMBIENT_COLOUR, &colour);

	my_settings->get_bool(DEBUG_SHOW_NORMALS, &show_normals);

	GLfloat light_ambient[] = { colour.r, colour.g, colour.b, 1.0  };

	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, light_ambient);



	my_settings->get_colour(DIRECT_COLOUR, &colour);
	GLfloat light_diffuse[] = { colour.r, colour.g, colour.b, 1.0  };
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);


	glDisable(GL_CULL_FACE);


	bool in_strip = false;

	if (draw_land) {
		for (int y = 0; y < (map_width-1) ; y++) {

			// Vertex offset for start of this line.
			int line_offset = y*map_width;

			for (int x = line_offset + 1; x < line_offset + map_width; x++) {

				// We're clipping in eye space, so we need to transform every
				//  point <sigh>  It would make a lot more sense to transform
				//  the viewing frustrum and clip to that
				point3d p1 = land.points[x];
				point3d p2 = land.points[x + map_width];

				p1 = camera*p1;
				p2 = camera*p2;

				bool visible1 = view_frustrum.test_point(p1);
				bool visible2 = view_frustrum.test_point(p2);

				if (visible1 || visible2) {
					if (!in_strip) {
						// If we're not in a strip, start one and add current and previous points.
						glBegin(GL_TRIANGLE_STRIP);
						add_point(land, x-1);
						add_point(land, x+map_width-1);

						add_point(land, x);
						add_point(land, x+map_width);
						in_strip = true;
					} else {
						add_point(land, x);
						add_point(land, x+map_width);
					}
				} else if (visible1 && visible2) {
					if (in_strip) {
						// This part not visible, so end triangle strip
						add_point(land, x);
						add_point(land, x+map_width);
						glColor3f(1.0, 1.0, 0.0);
						glEnd();
						in_strip = false;
					}
				}

			}

			// Finish this triangle strip if we were rendering
			if (in_strip) {
				in_strip = false;
				glEnd();
			}

		}
	}

	in_strip = false;


	if (draw_sky) {
		for (int y = 0; y < (map_width-1) ; y++) {

			int line_offset = y*map_width;

			in_strip = false;

			for (int x = line_offset + 1; x < line_offset + map_width; x++) {

				point3d p1 = sky.points[x];
				point3d p2 = sky.points[x + map_width];

				p1 = camera*p1;
				p2 = camera*p2;


				bool visible = view_frustrum.test_point(p1) && view_frustrum.test_point(p2);

				if (visible) {
					if (!in_strip) {
						// If we're not in a strip, start one and add current and previous points.
						glBegin(GL_TRIANGLE_STRIP);
						add_point(sky, x-1);
						add_point(sky, x+map_width-1);

						add_point(sky, x);
						add_point(sky, x+map_width);
						in_strip = true;
					} else {
						add_point(sky, x);
						add_point(sky, x+map_width);
					}
				} else {
					if (in_strip) {
						// This part not visible, so end triangle strip
						add_point(sky, x);
						add_point(sky, x+map_width);
						glColor3f(1.0, 0.0, 0.0);
						glEnd();
						in_strip = false;
					}
				}

			}
			// Finish this triangle strip if we were rendering
			if (in_strip) {
				in_strip = false;
				glEnd();
			}

		//break;
		}
	}


	// Display surface normals if requested
	if (show_normals) {
		draw_normals();
	}

	glPopMatrix();
}


/**
 * Execute GL commands for given vertex details
 */
void base_GLwindow::add_point(const vector3d& normal, const point3d& vertex, const fcolour& colour)
{
	glColor3f(colour.r, colour.g, colour.b);

	float col[4] = { 0, 0, 0, 1 };
	col[0] = colour.r; col[1] = colour.g; col[2] = colour.b;

	//glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, col);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, col);
	glNormal3f(normal.i, normal.j, normal.k);
	glVertex3f(vertex.x, vertex.y, vertex.z);
}


void base_GLwindow::add_point(object& obj, int idx) const
{
	fcolour colour = obj.points[idx].get_colour();
	glColor3f(colour.r, colour.g, colour.b);

	float col[4] = { 0, 0, 0, 1 };
	col[0] = colour.r; col[1] = colour.g; col[2] = colour.b;

	vector3d normal = obj.normals[idx];
	point3d vertex = obj.points[idx];

	//glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, col);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, col);
	glNormal3f(normal.i, normal.j, normal.k);
	glVertex3f(vertex.x, vertex.y, vertex.z);

}


/**
 * Draw some of the surface normals for debugging purposes
 */
void base_GLwindow::draw_normals(void)
{
	float col[4] = { 0, 0, 0, 1 };

	glColor3f(1.0, 1.0, 1.0);
	col[0] = 256.0/256.0; col[1] = 256.0/256.0; col[2] = 256.0/256.0;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, col);

	for (int i = 0; i < (map_width-1)*(map_width-1); i++) {

		if (i%4 || (i/map_width)%4) {
			continue;
		}

		glBegin(GL_LINES);

		point3d pnt(land.points[i].x, land.points[i].y, land.points[i].z);

		vector3d vec(land.normals[i].i, land.normals[i].j, land.normals[i].k);
		vec.normalise();
		vec = vec * 5;

		glVertex3f(pnt.x, pnt.y, pnt.z);

		pnt = pnt + vec;

		glVertex3f(pnt.x, pnt.y, pnt.z);

		glEnd();
	}
}


/**
 * Render model in wireframe
 */
void base_GLwindow::wireframe(void)
{
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
}


/**
 * Use simple flat shading
 */
void base_GLwindow::flat(void)
{
	glPolygonMode(GL_FRONT, GL_FILL );
	glShadeModel(GL_FLAT);
}


/**
 * Enable Gouraud shading for polygons
 */
void base_GLwindow::smooth(void)
{
	glPolygonMode(GL_FRONT, GL_FILL);
	glShadeModel(GL_SMOOTH);
}


/**
 * Simple Lambertian lighting
 */
void base_GLwindow::set_lighting(bool state)
{
	(state) ? glEnable(GL_LIGHTING)
	        : glDisable(GL_LIGHTING);
}


/**
 * Enable/disable fog for depth queueing
 */
void base_GLwindow::set_fogging(bool state)
{
	if (state) {
		glClearColor(fog_colour[0], fog_colour[1], fog_colour[2], 1.0);
		glEnable(GL_FOG);
	} else {
		glClearColor(sky_colour[0], sky_colour[1], sky_colour[2], 1.0);
		glDisable(GL_FOG);
	}
}


/**
 * Render model with shadows
 * @todo Fudge some shadow code.
 */
void base_GLwindow::set_shadows(bool /*state*/)
{
	// Unimplemented
}


/**
 * Set direction of infinite light source
 */
void base_GLwindow::set_light_position(float i, float j, float k)
{
	i = -i;
	j = -j;
	k = -k;

	light_i = i;
	light_j = j;
	light_k = k;

	GLfloat light_position[] = { light_i, light_j, light_k, 0.0 };

	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
}


/**
 * Metric for render detail
 */
void base_GLwindow::set_detail(int val)
{
	map_width = 1<<val;
}


/**
 * Enable/disable texturing of rendered polygons
 */
void base_GLwindow::set_texture(bool state)
{
	texturing = state;
}


#if 0
/**
 * Set rotation of rendered model
 * @param a,b,c	axes roation in radians
 */
void base_GLwindow::set_rotation(float a, float b, float c)
{
	// Convert radians into degrees
	a = a/(2*M_PI) * 360.0;
	b = b/(2*M_PI) * 360.0;
	c = c/(2*M_PI) * 360.0;

	rot_a = a;
	rot_b = b;
	rot_c = c;

}


/**
 * Set translation of rendered model
 */
void base_GLwindow::set_translation(float x, float y, float z)
{
	pos_x = x;
	pos_y = y;
	pos_z = z;
}
#endif
