//
// Colour.cc
//
// J. Belson	2000.06.08
//


#include <cstdio>

#include <Application.h>
#include <Bitmap.h>
#include <Font.h>
#include <Window.h>

#include "Colour.h"

// Message for each slider
static const uint32	MSG_A	= 'sl-a';
static const uint32	MSG_B	= 'sl-b';
static const uint32	MSG_C	= 'sl-c';


Colour::Colour(
			BRect frame,
			ConvertColour* Conv,
			BMessage* message,
			uint32 mode,
			uint32 flags)
	: BView(frame, "colour", mode, flags),
	Convert(Conv),
	ColMsg(message)
{
	BFont font;
	struct font_height fh;
	GetFont(&font);
	font.GetHeight(&fh);
	font_height = fh.ascent /*+ fh.descent*/;

	BRect slider_rect = Bounds();
	slider_rect.left += 14.0;

	float height = slider_rect.Height() - 1.0;
	float width = slider_rect.Width() - SQR_WIDTH;
	float division = height/3.0;		// Space per slider

	col Black	= { 0.0, 0.0, 0.0 };
	col Red		= { 1.0, 0.0, 0.0 };
	col Green	= { 0.0, 1.0, 0.0 };
	col Blue	= { 0.0, 0.0, 1.0 };

	BRect square_rect = BRect(width + 2, division/4, slider_rect.Width(), height);
	square_rect.OffsetBy( font_height /*14.0*/, 0.0);
	square = new BView(square_rect, "square", 0, 0);
	AddChild(square);

	BRect r(0, 0.0, width, division);
	r.OffsetBy( font_height /*14.0*/, 0.0);
	r.top += division/4;

	// Create the sliders
	slider[0] = new Slider(r, "slider1", "R", Convert, new BMessage(MSG_A));
	slider[0]->SetBarColour(Black, Red);

	r.OffsetBy(0.0, division);
	slider[1] = new Slider(r, "slider2", "G", Convert, new BMessage(MSG_B));
	slider[1]->SetBarColour(Black, Green);

	r.OffsetBy(0.0, division);
	slider[2] = new Slider(r, "slider3", "B", Convert, new BMessage(MSG_C));
	slider[2]->SetBarColour(Black, Blue);


	AddChild(slider[0]);
	AddChild(slider[1]);
	AddChild(slider[2]);

	// Add default values
	ColMsg->AddInt8("red", 0);
	ColMsg->AddInt8("green", 0);
	ColMsg->AddInt8("blue", 0);
}


Colour::~Colour()
{
	for (int i=0; i<3; i++) {
		delete slider[i];
	}
}



// Change the colourspace by supplying new converter
void Colour::SetConverter(ConvertColour* Conv)
{
	// Convert current colour to RGB
	uint8 red, green, blue;
	Convert->ToRGB(colour, &red, &green, &blue);

	Convert = Conv;

	// Find new slider positions for current colour
	colour = Convert->FromRGB(red, green, blue);


	slider[0]->Set(colour.a);
	slider[1]->Set(colour.b);
	slider[2]->Set(colour.c);

	// Tell slider to use our new colour model
	for (int i=0; i<3; i++) {
		slider[i]->SetConverter(Conv);
	}

	DrawSquare();
	DrawLabels();
	UpdateSliders();
}



// Set widget's current colour
void Colour::SetColour(uint8 red, uint8 green, uint8 blue)
{
	// Find new slider positions for current colour
	colour = Convert->FromRGB(red, green, blue);

	slider[0]->Set(colour.a);
	slider[1]->Set(colour.b);
	slider[2]->Set(colour.c);

	DrawSquare();
	DrawLabels();
	UpdateSliders();
}



void Colour::MessageReceived(BMessage* msg)
{
	switch (msg->what) {
		case MSG_A:
			colour.a = slider[0]->GetValue();
			DrawSquare();
			UpdateSliders();
			SendRGB();
			break;

		case MSG_B:
			colour.b = slider[1]->GetValue();
			DrawSquare();
			UpdateSliders();
			SendRGB();
			break;

		case MSG_C:
			colour.c = slider[2]->GetValue();
			DrawSquare();
			UpdateSliders();
			SendRGB();
			break;

		default:
			BView::MessageReceived(msg);
			break;
	}
}



//
// Update the colour gardients for each slider
//
void Colour::UpdateSliders(void)
{
	col left, right;

	left = colour;
	right = colour;
	left.a = 0.0;
	right.a = 1.0;
	slider[0]->SetBarColour(left, right);

	left = colour;
	right = colour;
	left.b = 0.0;
	right.b = 1.0;
	slider[1]->SetBarColour(left, right);

	left = colour;
	right = colour;
	left.c = 0.0;
	right.c = 1.0;
	slider[2]->SetBarColour(left, right);
}



// Send updated RGB values to parent window
void Colour::SendRGB(void)
{
	ColMsg->ReplaceInt8("red",   255.0*slider[0]->GetValue());
	ColMsg->ReplaceInt8("green", 255.0*slider[1]->GetValue());
	ColMsg->ReplaceInt8("blue",  255.0*slider[2]->GetValue());

	Window()->PostMessage(ColMsg);
}



void Colour::AttachedToWindow(void)
{
	// Inherit view colour from our parent
	if (Parent()) {
		SetViewColor(Parent()->ViewColor());
		SetLowColor(ViewColor());
	}
}



//
// Draw mixer square in current colour
//
void Colour::DrawSquare(void)
{
	BRect r = square->Bounds();

	square->SetHighColor(0, 0, 0);
	square->StrokeRect(r);
	r.InsetBy(1.0, 1.0);

	uint8 red, green, blue;
	Convert->ToRGB(colour, &red, &green, &blue);

	square->SetHighColor(red, green, blue);
	square->FillRect(r);
}



//
// Draw slider labels
//
void Colour::DrawLabels(void)
{
	FillRect(BRect(0.0, Bounds().top, slider[0]->Frame().left-1, Bounds().bottom), B_SOLID_LOW);

	// Draw slider labels at strategic points
	for (int i=0; i<3; i++) {
		BRect slider_frame = slider[i]->Frame();

		float offset = (slider_frame.Height()-5)/2; /*- font_height*/;
		BPoint p(0.0, slider_frame.bottom - offset + font_height/2.0 /*- font_height*/);

		DrawChar(Convert->letter[i], p);
	}
}



void Colour::Draw(BRect update)
{
	DrawSquare();
	DrawLabels();
}



