//
// mouse_bm_view.cc
//
// View that responds to mouse clicks
//
// Copyright (c) J. Belson	1998.12.12
//

#include <iostream.h>

#include "matrix.h"
#include "mouse_bm_view.h"
#include "vector3d.h"


//
// mouse_bm_view
//
mouse_bm_view::mouse_bm_view(BRect rect)
	: BView(rect, "relief view", 0, B_WILL_DRAW),
	  tracking(TRACKING_NONE)
{
	bm			= new BBitmap(rect, B_RGB32 /*B_CMAP8*/);
	bm_data		= (uchar *) bm->Bits();
	bytes_per_row	= bm->BytesPerRow();

	WIDTH	= int (rect.right - rect.left);

	reset_camera();
}



//
// Move camera to default position and orientation
//
void mouse_bm_view::reset_camera(void)
{
	camera_x = WIDTH/2;
	camera_y = WIDTH - 1;
	target_x = WIDTH/2;
	target_y = 0;
	
	Draw(Bounds());
}



void mouse_bm_view::MouseDown(BPoint point)
{
	int32 buttons = 0;

	Window()->CurrentMessage()->FindInt32((const char *) "buttons", &buttons);


	if (buttons & B_PRIMARY_MOUSE_BUTTON) {
		tracking = TRACKING_CAMERA;
		update(point);
	} else if (buttons & B_SECONDARY_MOUSE_BUTTON) {
		tracking = TRACKING_TARGET;
		update(point);
	}

}



//
// Camera or target has moved, so update and inform
//  parent window
//
void mouse_bm_view::update(BPoint point)
{
	BMessage msg;

	draw_line_of_sight();

	switch (tracking) {
		case TRACKING_CAMERA:
			camera_x = int (point.x);
			camera_y = int (point.y);
			msg.what = MSG_NEW_CAMERA;
			msg.AddInt32("x", camera_x);
			msg.AddInt32("y", camera_y);
			break;

		case TRACKING_TARGET:
			target_x = int (point.x);
			target_y = int (point.y);
			msg.what = MSG_NEW_TARGET;
			msg.AddInt32("x", target_x);
			msg.AddInt32("y", target_y);
			break;

		default:
			;
	}
	Window()->PostMessage(&msg);

	draw_line_of_sight();
}



//
// If mouse moves out of the BView, leave tracking mode
//
void mouse_bm_view::MouseUp(BPoint point)
{
	update(point);
	tracking = TRACKING_NONE;
}



//
// Draw line from camera to target
//
void mouse_bm_view::draw_line_of_sight(void)
{
	const float fov = 0.4;


	SetDrawingMode(B_OP_INVERT);

	// Draw left/right bounds of line of sight...
	vector3d v2(target_x - camera_x, 0.0, target_y - camera_y);

	vector3d lh = matrix(matrix::ROTATION, 0, -fov, 0)*v2;
	StrokeLine(BPoint(camera_x, camera_y), BPoint(camera_x + lh.i, camera_y + lh.k));

	vector3d rh = matrix(matrix::ROTATION, 0, fov, 0)*v2;
	StrokeLine(BPoint(camera_x, camera_y), BPoint(camera_x + rh.i, camera_y + rh.k));


	// Draw a cross at the target...
	matrix m = matrix(matrix::ROTATION, 0, M_PI/2.0, 0);

	vector3d v(target_x - camera_x, 0.0, target_y - camera_y, true);
	v = v*10;
	v = matrix(matrix::ROTATION, 0, M_PI/4.0, 0)*v;

	for (int i=0; i<4; i++) {
		StrokeLine(BPoint(target_x, target_y), BPoint(target_x + v.i, target_y + v.k));
		v = m*v;
	}


	SetDrawingMode(B_OP_COPY);
}



//
// Will use this for real-time preview updating...
//
void mouse_bm_view::MouseMoved(BPoint point, uint32 transit, const BMessage *message)
{
	switch (transit) {
		case B_INSIDE_VIEW:
			switch (tracking) {
				case TRACKING_CAMERA:
					update(point);
					break;
				case TRACKING_TARGET:
					update(point);
					break;
				default:
					;
			}
			break;

		case B_ENTERED_VIEW:
			//cout << "Mouse entered" << endl;
			break;

		case B_EXITED_VIEW:
			//cout << "Mouse left" << endl;
			tracking = TRACKING_NONE;
			break;
	}
}



void mouse_bm_view::get_camera_pos(int *x, int *y)
{
	*x = camera_x;
	*y = camera_y;
}


void mouse_bm_view::set_camera_pos(int x, int y)
{
	draw_line_of_sight();
	camera_x = x;
	camera_y = y;
	
	cout << "set_camera_pos(" << x << ", " << y << ")\n";

	draw_line_of_sight();

	Draw(Bounds());
}


void mouse_bm_view::get_target_pos(int *x, int *y)
{
	*x = target_x;
	*y = target_y;
}

void mouse_bm_view::set_target_pos(int x, int y)
{
	draw_line_of_sight();
	target_x = x;
	target_y = y;

	draw_line_of_sight();

	Draw(Bounds());
}





