//
// preview_window.cc
//
// Adjustable preview window
//
// Copyright (c) J. Belson	1998.12.14
//


#include <iostream.h>
#include <Bitmap.h>
#include <Screen.h>
#include <Message.h>

#include "MApplication.h"
#include "VGroup.h"
#include "HGroup.h"
#include "MButton.h"
#include "MBViewWrapper.h"
#include "MBorder.h"
#include "MPopup.h"
#include "PropGadget.h"
#include "MCheckBox.h"
#include "MRadioGroup.h"
#include "Space.h"

#include "hf_fbm.h"
#include "hf_synthesis.h"
#include "hf_perlin.h"
#include "messages.h"
#include "not_implemented.h"
#include "preview_window.h"
#include "settings.h"



//
// Constructor/destructor
//
preview_window::preview_window(
		BWindow *window, BRect r, char *name, window_type t, ulong mode)
	: smart_window(r, name, t, mode),
	  parent(window)
{
	// Put output window next to control window
	BRect rect = r;
	rect.InsetBy(r.right-r.left + 140, 0);
	rect.bottom	= rect.top + w_height; 
	rect.right	= rect.left + w_width;

	win = new GLwindow(rect, "GL Preview Output",
						/*B_FLOATING_WINDOW*/	 B_TITLED_WINDOW,
						B_NOT_RESIZABLE | B_NOT_CLOSABLE | B_NOT_ZOOMABLE);
	win->GLinit();


	MCheckBox*	cb_smoothing;
	MCheckBox*	cb_lighting;
	MCheckBox*	cb_normals;
	MCheckBox*	cb_texture;
	MCheckBox*	cb_fogging;
	MCheckBox*	cb_shadows;
	MCheckBox*	cb_update;
	MPopup*		pop_detail;

	cb_smoothing= new MCheckBox("Smoothing", new BMessage(MSG_SMOOTHING), NULL, smoothing);
	cb_smoothing->SetEnabled(false);
	cb_normals	= new MCheckBox("Show normals", new BMessage(MSG_NORMALS), NULL, normals);
	cb_lighting	= new MCheckBox("Lighting", new BMessage(MSG_LIGHTING), NULL, lighting);
	cb_fogging	= new MCheckBox("Fogging", new BMessage(MSG_FOGGING), NULL, fogging);
	cb_texture	= new MCheckBox("Texture", new BMessage(MSG_TEXTURE), NULL, texture);
	cb_shadows	= new MCheckBox("Shadows", new BMessage(MSG_FOGGING), NULL, shadows);
	cb_update	= new MCheckBox("Auto update", new BMessage(MSG_UPDATE), NULL, auto_update);
	pop_detail 	= new MPopup("Detail", new BMessage(MSG_DETAIL), this, "Low", "Medium", "High", 0);


	MView *top_view =
		new VGroup(
			new Space(minimax(4, 4, 4, 4)),
			new MBorder(
				M_LABELED_BORDER, 12, "Options",
				new VGroup (
					cb_smoothing,
					//cb_normals,
					cb_lighting,
					cb_fogging,
					cb_texture,
					cb_shadows,
					0
				)
			),
			new MBorder(
				M_LABELED_BORDER, 12, "Display",
				new VGroup(
					cb_update,
					pop_detail,
					0
				)
			),

			new MButton("Update", new BMessage(MSG_RERENDER), 0, minimax(-1,-1,-1,-1)),
			new Space(minimax(4, 4, 4, 4)),
			0
		);

	Lock();
	// top_view is an MView, so we need to cast it to a BView.
	AddChild(dynamic_cast<BView*> (top_view));
	Unlock();

	pop_detail->SetActive(1);
	pop_detail->SetAlignment(B_ALIGN_CENTER);

	cb_texture->SetEnabled(false);
	//cb_fogging->SetEnabled(false);
	cb_shadows->SetEnabled(false);

	the_thread	= 0;
}



preview_window::~preview_window()
{
	BMessage msg(MSG_PREVIEW_CLOSED);
	parent->PostMessage(&msg);
}



//
// Deal with window close gadget
//
bool preview_window::QuitRequested()
{
	smart_window::QuitRequested();

	if (win->Lock()) {
		cout << "calling win::Quit()" << endl;
		win->Quit();
		//win->PostMessage(new BMessage(B_QUIT_REQUESTED));
	} else {
		cerr << "Lock failed!!!" << endl;
	}

	return TRUE;
}



//
// Deal with incoming messages to window
//
void preview_window::MessageReceived(BMessage *message)
{
	switch (message->what)
	{
		//
		// Local messages
		//
		case MSG_LIGHTING:
			lighting = !lighting;
			win->set_lighting(lighting);
			break;

		case MSG_RERENDER:
			if (the_thread) {
				//cerr << "Can't rerender, I'm already rendering!" << endl;
			} else {
				the_thread = spawn_thread(/*rerender_*/ preview_thread_func, "Rerender Thread",
						B_NORMAL_PRIORITY, this);
				resume_thread(the_thread);
			}
			break;

		case MSG_SMOOTHING:
			smoothing = !smoothing;
			(smoothing)	? win->smooth()
						: win->flat();
			break;

		case MSG_UPDATE:
			auto_update = !auto_update;
			break;

		case MSG_DETAIL:
			{
			int32 level;
			if (message->FindInt32("index", &level) == B_NO_ERROR) {
				detail = 6 + level;
				need_recalc = true;
			}
			break;
			}

		case MSG_TEXTURE:
			break;

		case MSG_FOGGING:
			fogging = !fogging;
			win->set_fogging(fogging);
			break;

		case MSG_SHADOWS:
			break;

		case MSG_NORMALS:
			normals = !normals;
			break;

		//
		// Global messages
		//
		case MSG_NEW_MAP:
			//cout << "preview got MSG_UPDATE_MAP" << endl;

			need_recalc = true;

			my_settings.get_int(PREVIEW_SEED, &seed);
			my_settings.get_float(SYNTHESIS_FD, &fd);

			if (the_thread == 0 && auto_update) {
				the_thread = spawn_thread(preview_thread_func, "Preview Thread",
						B_NORMAL_PRIORITY, this);
				resume_thread(the_thread);
			} else {
				//cerr << "Rendering thread already running!!" << endl;
			}
			break;


		default:
			MWindow::MessageReceived(message);
			break;
	}
}



//
// Static function to launch thread
//
long preview_window::preview_thread_func(void *obj)
{
	preview_window *my_obj = (preview_window *) obj;
	my_obj->render_thread();
	my_obj->the_thread = 0;
	return 0;
}



long preview_window::rerender_thread_func(void *obj)
{
	preview_window *my_obj = (preview_window *) obj;
	my_obj->rerender_thread();
	return 0;
}



void preview_window::rerender_thread(void)
{
	win->rerender();
	the_thread = 0;
}







