//
// sunview.cc
//
// Allow user to position sun
//
// Copyright (c) J. Belson	2000.05.14
//

#include <iostream>

#include "settings.h"
#include "sunview.h"
#include "vector3d.h"


above_view::above_view(BMessage* message)
	: BView(BRect(0, 0, 128, 128), "above_view", 0, B_WILL_DRAW),
	  msg(message),
	  MAX_SIZE(128)
{
	centre_x = MAX_SIZE/2.0;
	centre_y = MAX_SIZE/2.0;

	vector = vector2d(0.0, -MAX_SIZE);
	
	tracking = false;
}

void above_view::MouseDown(BPoint point)
{
	start_tracking();
	update(point.x, point.y);
}


void above_view::MouseUp(BPoint point)
{
	stop_tracking();
}


void above_view::MouseMoved(
						BPoint point,
						uint32 transit,
						const BMessage *message)
{
	if (tracking) {
		update(point.x, point.y);
	}
}

void above_view::start_tracking(void)
{
	tracking = true;
}



void above_view::stop_tracking(void)
{
	tracking = false;

//	if (B_NAME_NOT_FOUND == msg->ReplaceFloat("angle", angle)) {
//		msg->AddFloat("angle", angle);
//	}

	Window()->PostMessage(msg);

	MessageReceived(msg);
}

void above_view::update(float x, float y)
{
	// Make vector between original and (x, y)
	vector2d v(x - centre_x, y - centre_y);
	v.normalise();
	v = v*MAX_SIZE;

	vector = v;
	
	Draw(Bounds());
}

void above_view::Draw(BRect rect)
{
	SetHighColor(40, 166, 40);

	FillRect(Bounds());
	
	SetHighColor(0, 0, 0);
	FillRect(BRect(centre_x-1.0, centre_y-1.0, centre_x+1.0, centre_y+1.0));
	StrokeLine( BPoint(centre_x, centre_y),
				BPoint(centre_x+vector.i, centre_y+vector.j));
}

float above_view::calculate_angle(void)
{
	vector2d reference(0.0, -1.0);
	float cos_angle = reference.dot_product(vector) / (reference.magnitude()*vector.magnitude());
	float angle = acos(cos_angle);

	if (vector.i < 0.0) {
		angle = 2*M_PI - angle;
	}

	return angle;
}




side_view::side_view(BMessage* message)
	: BView(BRect(0, 0, 128, 128), "side_view", 0, B_WILL_DRAW),
	  msg(message),
	  MAX_SIZE(128)
{
	centre_x	= 1.0;
	centre_y	= MAX_SIZE/2.0;

	vector = vector2d(0.0, -MAX_SIZE/2.0);
	
	tracking = false;
}

void side_view::MouseMoved(
						BPoint point,
						uint32 transit,
						const BMessage *message)
{
	if (tracking) {
		int x = int (point.x);
	
		if (x < 1) {
			x = 1;
		}
		update(x, point.y);
	}
}

void side_view::Draw(BRect rect)
{
	SetHighColor(82, 134, 214);
	FillRect(BRect(0, 0, MAX_SIZE, MAX_SIZE/2.0));
	SetHighColor(40, 166, 40);
	FillRect(BRect(0, MAX_SIZE/2.0, MAX_SIZE, MAX_SIZE));

	SetHighColor(0, 0, 0);
	FillRect(BRect(centre_x-1.0, centre_y-1.0, centre_x+1.0, centre_y+1.0));
	StrokeLine( BPoint(centre_x, centre_y),
				BPoint(centre_x+vector.i, centre_y+vector.j));
}

void side_view::update(float x, float y)
{
	// Make vector between original and (x, y)
	vector2d v(x - centre_x, y - centre_y);
	v.normalise();
	v = v*MAX_SIZE;

	vector = v;
	
	Draw(Bounds());
}

//
// Calculate angle of lift
//
float side_view::calculate_angle(void)
{
	vector2d reference(0.0, -1.0);

	float cos_angle = reference.dot_product(vector) / (reference.magnitude()*vector.magnitude());
	float angle = acos(cos_angle);

	if (vector.i < 0.0) {
		angle = -angle;
	}

	// Convert scale from (0 -> PI) to (PI -> -PI)
	angle = -angle + M_PI/2.0;

angle = -angle;

	return angle;
}

void side_view::MouseDown(BPoint point)
{
	start_tracking();
	update(point.x, point.y);
}


void side_view::MouseUp(BPoint point)
{
	stop_tracking();
}

void side_view::start_tracking(void)
{
	tracking = true;
}

void side_view::stop_tracking(void)
{
	tracking = false;

	Window()->PostMessage(msg);

	MessageReceived(msg);
}







sunview::sunview(BRect rect, BMessage* message)
	: BView(rect, "sun view", 0, B_WILL_DRAW)
{
	msg = message;

	width	= (rect.right - rect.left);
	height	= (rect.bottom - rect.top);

	x = width/2.0;
	y = height/2.0;

	tracking = false;

}



void sunview::set_position(float i, float j, float k)
{
	x = width/2 - i;
	y = width/2 - k;

	Draw(Bounds());
}



bool sunview::in_range(BPoint p)
{
	bool ret_val = false;

	if (p.x > x - RADIUS &&
		p.x < x + RADIUS &&
		p.y > y - RADIUS &&
		p.y < y + RADIUS) {

		ret_val = true;
	}

	return ret_val;
}


void sunview::start_tracking(void)
{
	tracking = true;
}



void sunview::stop_tracking(void)
{
	tracking = false;

	float i = width/2 - x;
	float j = -width/2;
	float k = width/2 - y;

	float magnitude = sqrt(i*i + j*j + k*k);
	if (magnitude) {
		i /= magnitude;
		j /= magnitude;
		k /= magnitude;
	}

	// Send back updated coords
	if (B_NAME_NOT_FOUND == msg->ReplaceFloat("i", i)) {
		msg->AddFloat("i", i);
	}
	if (B_NAME_NOT_FOUND == msg->ReplaceFloat("j", j)) {
		msg->AddFloat("j", j);
	}
	if (B_NAME_NOT_FOUND == msg->ReplaceFloat("k", k)) {
		msg->AddFloat("k", k);
	}

	Window()->PostMessage(msg);

	MessageReceived(msg);
}


void sunview::draw_sun(void)
{
	StrokeEllipse(BRect(x - RADIUS, y - RADIUS, x + RADIUS, y + RADIUS));
}



void sunview::Draw(BRect rect)
{
	draw_sun();

}

void sunview::MouseDown(BPoint point)
{
	if (in_range(point)) {
		start_tracking();
	}
}

void sunview::MouseMoved(BPoint point, uint32 transit, const BMessage *message)
{
	if (tracking /*&& transit == B_INSIDE_VIEW*/) {

		if (point.x > 0 &&
			point.x < width &&
			point.y > 0 &&
			point.y < height) {

			SetDrawingMode(B_OP_INVERT);
			draw_sun();
	
			x = point.x;
			y = point.y;
	
			SetDrawingMode(B_OP_COPY);
			draw_sun();
		}
	}
}

void sunview::MouseUp(BPoint point)
{
	stop_tracking();
}



