//
// bm_window.cc
//
// Copyright (C) J. Belson	2000.03.20
//
// $Author: jon $	:	$Date: 2003/05/29 14:45:06 $
//

#include <qpainter.h>
#include <qstring.h>

#include <cstdio>
#include <cstring>
#include <cerrno>
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>

#include "bm_window.h"


using namespace std;


/**
 * Constructor
 */
bm_window::bm_window(int w, int h, QWidget* parent, const char* name)
	: QWidget(parent, name, WStyle_Customize|WStyle_DialogBorder),
	  image_full(0),
	  image_scaled(0),
	  render_buffer(0),
	  smooth_update(false)
{
	setFixedSize(w, h);
	setBackgroundColor(white);

	set_size(w, h);

	redraw_timer = new QTimer(this);
	connect(redraw_timer, SIGNAL(timeout()), this, SLOT(slot_refresh()));
}


/**
 * Destructor
 */
bm_window::~bm_window()
{
	delete image_full;
	delete image_scaled;

	if (render_buffer) {
		munmap(render_buffer, render_buffer_size);
	}
}


/**
 * Enable/disable antialiased repainting
 */
void bm_window::set_smooth_update(bool b)
{
	smooth_update = b;
}


/**
 * Get pointer to bitmap data
 */
uint32* bm_window::get_buffer(void)
{
	#ifdef USE_MMAP
	return (uint32*) render_buffer;
	#else 
	return (uint32*) image_full->bits();
	#endif
}


/**
 * Get bytes per row for bitmap data
 */
int bm_window::bytes_per_row(void)
{
	return image_full->bytesPerLine();
}



/**
 * Repaint window from current render buffer
 */
void bm_window::slot_refresh(void)
{
	QPainter p(this);

	if (smooth_update) {
		#ifdef USE_MMAP
		// Copy mmap()ed render_buffer to image_full
		for (int y=0; y<render_height; y++) {
			uint32* p = (uint32*) (render_buffer);
			p += y*bytes_per_row()/4;
			uint32* q = (uint32*) image_full->scanLine(y);

			for (int x=0; x<render_width; x++) {
				*q++ = *p++;
			}
		}
		#endif

		*image_scaled = image_full->smoothScale(width, height);

	} else {

#if 0
		for (int y=0; y<height; y++) {
			#ifdef USE_MMAP
			uint32* p = (uint32*) (render_buffer);
			p += SCALE*y*bytes_per_row()/4;
			#else
			uint32* p = (uint32*) image_full->scanLine(y*SCALE);
			#endif
			uint32* q = (uint32*) image_scaled->scanLine(y);
			for (int x=0; x<width; x++) {
				*q++ = *p;
				p += SCALE;		// Skip alternate pixels
			}
		}
#endif
		*image_scaled = image_full->scale(width, height);

	}

	p.drawImage(0, 0, *image_scaled);

}


/**
 * Set window to required size
 * @param w New window width
 * @param h	New window height
 */
void bm_window::set_size(int w, int h)
{
	setFixedSize(w, h);

	width = w;
	height = h;

	render_width = width*SCALE;
	render_height = height*SCALE;

printf("set_size(%dx%d) (%dx%d)\n", width, height, render_width, render_height);

	if (image_full) {
		delete image_full;
	}
	image_full = new QImage(render_width, render_height, 32);

	if (image_scaled) {		// Shouldn't really need to check both explicitly
		delete image_scaled;
	}
	image_scaled = new QImage(width, height, 32);

	#ifdef USE_MMAP
	if (render_buffer) {
		int status = munmap(render_buffer, render_buffer_size);
		if (status == -1) {
			printf("munmap() failed - not good\n");
		} else {
			puts("munmap() succeeded");
		}
	}

	render_buffer_size = bytes_per_row()*render_height;
	render_buffer = mmap(0, render_buffer_size, PROT_READ|PROT_WRITE, MAP_ANON|MAP_SHARED, -1, 0);
	if (render_buffer == MAP_FAILED) {
		puts("mmap() failed");
		printf("errno = %d\n", errno);
		render_buffer = 0;
	} else {
		puts("mmap() succeeded");
	}
	#endif

	clear_window();
}


/**
 * Clear image buffer to black
 */
void bm_window::clear_window(void)
{
	for (int y=0; y<render_height; y++) {
		uint32* p = (uint32*) image_full->scanLine(y);
		for (int x=0; x<render_width; x++) {
			*p++ = 0;
		}
	}

	for (int y=0; y<height; y++) {
		uint32* p = (uint32*) image_scaled->scanLine(y);
		for (int x=0; x<width; x++) {
			*p++ = 0;
		}
	}
	#ifdef USE_MMAP
	std::memset(render_buffer, 0, render_buffer_size);
	#endif
}


/**
 * Rate at which window contents is updated
 * @param rate Redraw rate in pulses per sec, 0 stops updating
 */
void bm_window::set_redraw_rate(int rate)
{
	if (rate) {
		redraw_timer->start(int(1.0/rate * 1000));
	} else {
		redraw_timer->stop();
	}
}


/**
 * Get window width and height for clipping purposes or whatever
 */
void bm_window::get_window_size(int *x, int *y)
{
	*x = width;
	*y = height;
}


/**
 * Get width and height of area to be rendered in to
 */
void bm_window::get_render_size(int *x, int *y)
{
	*x = render_width;
	*y = render_height;
}


/**
 * Qt paint event
 */
void bm_window::paintEvent(QPaintEvent* /*e*/)
{
	QPainter p(this);

	if (smooth_update) {
		*image_scaled = image_full->smoothScale(width, height);
		p.drawImage(0, 0, *image_scaled);
	} else {
		*image_scaled = image_full->scale(width, height);	// ???
		p.drawImage(0, 0, *image_scaled);
	}
}


/**
 * Qt close event.  Unused at present
 */
void bm_window::closeEvent(QCloseEvent* /*e*/)
{
	;
}


/**
 * Get structure representing rendered image
 */
const QImage* bm_window::get_bitmap(void)
{
	return const_cast<const QImage*>(image_scaled);
}




