//
// light_vector.cc
//
// Custom widget to specify light direction
//
// Copyright (C) J. Belson	2000.12.11
//
// $Author: jon $	:	$Date: 2003/05/30 11:04:04 $
//


#include <cstdio>
#include <cmath>

#include <qpainter.h>

#include "light_vector.h"
#include "vector3d.h"

using namespace std;


/**
 * Constructor
 */
above_view::above_view(QWidget* parent, const char* name)
	: QWidget(parent, name),
	  MAX_SIZE(128)
{
	centre_x	= width()/2;
	centre_y	= height()/2;

	vector = vector2d(0.0, -MAX_SIZE);

	setMaximumSize(MAX_SIZE, MAX_SIZE);
}


/**
 * User moved mouse over widget, with at least one mouse button
 *  clicked
 */
void above_view::mouseMoveEvent(QMouseEvent* me)
{
	update(me->x(), me->y());
}


/**
 * User pressed a mouse button over widget
 */
void above_view::mousePressEvent(QMouseEvent* me)
{
	update(me->x(), me->y());
}


/**
 * Redraw light vector
 */
void above_view::paintEvent(QPaintEvent* /*pe*/)
{
	QPainter p(this);

	p.setBrush(QColor(40, 166,  40    /*0, 255, 0*/));
	p.setPen(NoPen);
	p.drawRect(0, 0, width(), height());
	p.setPen(SolidLine);

	p.setBrush(QColor(0, 0, 0));
	p.drawRect(centre_x-1, centre_y-1, 3, 3);
	p.drawLine(centre_x, centre_y, int(centre_x + vector.i), int(centre_y + vector.j));
}


/**
 * Widget resized, recalculate constants
 */
void above_view::resizeEvent(QResizeEvent* /*re*/)
{
	centre_x	= width()/2;
	centre_y	= height()/2;

	MAX_SIZE	= width()/2;
}


/**
 * Recalculate light vector for new position
 */
void above_view::update(float x, float y)
{
	// Make vector between original and (x, y)
	vector2d v(x - centre_x, y - centre_y);
	v.normalise();
	v = v*MAX_SIZE;

	vector = v;

	repaint(0, 0, width(), height(), /*true*/ false);
	
	send_signal();
}


/**
 * Calculate light vector rotation from -Y axis for above view
 */
float above_view::calculate_angle(void)
{
	vector2d reference(0.0, -1.0);
	float cos_angle = reference.dot_product(vector) / (reference.magnitude()*vector.magnitude());
	float angle = acos(cos_angle);

	if (vector.i < 0.0) {
		angle = 2*M_PI - angle;
	}

	return angle;
}


/**
 * Vector changed, notify anyone who is listening
 */
void above_view::send_signal(void)
{
	emit position_changed(); 
}



/**
 * Call private member to get rotation angle
 */
float above_view::get_rotation(void)
{
	return calculate_angle();
}



/**
 * Constructor
 */
side_view::side_view(QWidget* parent, const char* name)
	: QWidget(parent, name),
	  MAX_SIZE(128)
{
	centre_x	= 1.0;
	centre_y	= height()/2.0;
		
	vector = vector2d(0.0, -height()/2.0);
	setMaximumSize(MAX_SIZE, MAX_SIZE);
}


/**
 * User moved mouse over widget, with at least one mouse button
 *  clicked
 */
void side_view::mouseMoveEvent(QMouseEvent* me)
{
	int x = me->x();

	if (x < 1) {
		x = 1;
	}
	update(x, me->y());
}


/**
 * User pressed a mouse button over widget
 */
void side_view::mousePressEvent(QMouseEvent* me)
{
	update(me->x(), me->y());
}


/**
 * Redraw light vector
 */
void side_view::paintEvent(QPaintEvent* /*pe*/)
{
	QPainter p(this);

	p.setPen(NoPen);
	p.setBrush(QColor(82, 134, 214  /*0, 0, 255*/));
	p.drawRect(0, 0, width(), height()/2);
	p.setBrush(QColor(40,   166,  40  /*0, 255, 0*/));
	p.drawRect(0, height()/2, width(), height()/2);
	p.setPen(SolidLine);

	p.setBrush(QColor(0, 0, 0));
	p.drawRect(int (centre_x - 1), int (centre_y - 1), 3, 3);
	p.drawLine(int (centre_x), int (centre_y), int(centre_x + vector.i), int(centre_y + vector.j));
}


/**
 * Widget resized, recalculate constants
 */
void side_view::resizeEvent(QResizeEvent* /*re*/)
{
	vector = vector2d(0.0, -height()/2.0);
	centre_x	= 1;
	centre_y	= height()/2;
}



/**
 * Return angle of lift from horizontal
 */
float side_view::get_lift(void)
{
	return calculate_angle();
}



/**
 * Recalculate light vector for new position
 */
void side_view::update(float x, float y)
{
	// Make vector between original and (x, y)
	vector2d v(x - centre_x, y - centre_y);
	v.normalise();
	v = v*MAX_SIZE;

	vector = v;

	repaint(0, 0, width(), height(), /*true*/ false);

	send_signal();
}


/**
 * Calculate angle of lift for side view
 */
float side_view::calculate_angle(void)
{
	vector2d reference(0.0, -1.0);

	float cos_angle = reference.dot_product(vector)
				/ (reference.magnitude()*vector.magnitude());
	float angle = acos(cos_angle);

	if (vector.i < 0.0) {
		angle = -angle;
	}

	// Convert scale from (0 -> PI) to (PI -> -PI)
	angle = -angle + M_PI/2.0;

angle = -angle;

	return angle;
}


/**
 * Vector changed, notify anyone who is listening
 */
void side_view::send_signal(void)
{
	emit position_changed();
}


