//
// main_window.h
//
// Display and control the main application window
//
// Copyright (C) J. Belson
//
// $Author: jon $	:	$Date: 2003/09/17 20:59:35 $
//


#include <cstdio>
#include <ctime>

#include <klocale.h>
#include <kfiledialog.h>
#include <kimageio.h>
#include <kmessagebox.h>
#include <qlayout.h>
#include <qcombobox.h>
#include <qvbox.h>
#include <qvgroupbox.h>
#include <qlineedit.h>
#include <qmenubar.h>
#include <qmessagebox.h>
#include <qpainter.h>
#include <qwidgetstack.h>
#include <qspinbox.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qslider.h>
#include <qtabwidget.h>
#include <qvariant.h>
#include <qtooltip.h>
#include <qwhatsthis.h>


#include "atmospherics_window.h"
#include "colour.h"
#include "lighting_window.h"
#include "matrix.h"
#include "main.h"
#include "main_window.h"
#include "mouse_bm_view.h"
#include "point3d.h"
#include "preview_window.h"
#include "render_window.h"
#include "settings.h"
#include "settings_window.h"
#include "texture_window.h"
#include "texture.h"
#include "version.h"
#include "vector3d.h"
#include "widget_main.h"
#include "widget_multifractal.h"
#include "widget_perlin.h"
#include "widget_perlin2.h"
#include "widget_ridged_perlin.h"



/**
 * Constructor
 */
main_window::main_window(QWidget* parent)
	: widget_main(parent),
	  base_main_window(),
	  w_render(0),
	  w_preview(0),
	  w_settings(0)
{
	// Create settings widgets for available fractal types
	w_perlin		= new widget_perlin();
	w_perlin2		= new widget_perlin2();
	w_multifractal	= new widget_multifractal();
	w_ridged		= new widget_ridged_perlin();

	// Add generator types to popup menu
    comboBoxMethod->insertItem("Perlin");
    comboBoxMethod->insertItem("Perlin 2");
	comboBoxMethod->insertItem("Multifractal");
	comboBoxMethod->insertItem("Ridged Perlin");
	comboBoxMethod->setCurrentItem(get_type());

	stack->addWidget(w_perlin, 1);
	stack->addWidget(w_perlin2, 2);
	stack->addWidget(w_multifractal, 3);
	stack->addWidget(w_ridged, 4);
//	stack->raiseWidget(1 + get_type());
	slot_hf_type(get_type());

	update_widgets();

	// Make connections
	connect(pushButtonGenerate, SIGNAL(clicked()), this, SLOT(slot_generate()));
	connect(comboBoxMethod, SIGNAL(activated(int)), this, SLOT(slot_hf_type(int)));

	connect(w_perlin->spinBoxIterations, SIGNAL(valueChanged(int)), this, SLOT(slot_perlin_iterations(int)));
	connect(w_perlin->sliderScale, SIGNAL(valueChanged(int)), this, SLOT(slot_perlin_scale(int)));
	connect(w_perlin->sliderPersistence, SIGNAL(valueChanged(int)), this, SLOT(slot_perlin_persistence(int)));

	connect(w_perlin2->spinBoxOctaves, SIGNAL(valueChanged(int)), this, SLOT(slot_perlin2_iterations(int)));
	connect(w_perlin2->lineEditLacunarity, SIGNAL(returnPressed()), this, SLOT(slot_perlin2_lacunarity()));
	connect(w_perlin2->lineEditH, SIGNAL(returnPressed()), this, SLOT(slot_perlin2_h()));
	connect(w_perlin2->sliderScale, SIGNAL(valueChanged(int)), this, SLOT(slot_perlin2_scale(int)));

	connect(w_multifractal->spinBoxIterations, SIGNAL(valueChanged(int)), this, SLOT(slot_multifractal_iterations(int)));
	connect(w_multifractal->lineEditLacunarity, SIGNAL(returnPressed()), this, SLOT(slot_multifractal_lacunarity()));
	connect(w_multifractal->lineEditH, SIGNAL(returnPressed()), this, SLOT(slot_multifractal_h()));
	connect(w_multifractal->sliderScale, SIGNAL(valueChanged(int)), this, SLOT(slot_multifractal_scale(int)));

	connect(w_ridged->sliderScale, SIGNAL(valueChanged(int)), this, SLOT(slot_ridged_scale(int)));
	connect(w_ridged->spinBoxIterations, SIGNAL(valueChanged(int)), this, SLOT(slot_ridged_iterations(int)));
	connect(w_ridged->lineEditLacunarity, SIGNAL(returnPressed()), this, SLOT(slot_ridged_lacunarity()));
	connect(w_ridged->lineEditH, SIGNAL(returnPressed()), this, SLOT(slot_ridged_h()));
	connect(w_ridged->sliderGain, SIGNAL(valueChanged(int)), this, SLOT(slot_ridged_gain(int)));

	// Set up the relief view 
	relief_view->reset_camera();
	connect(relief_view, SIGNAL(signal_camera_moved()), this, SLOT(slot_camera_moved()));
	connect(relief_view, SIGNAL(signal_target_moved()), this, SLOT(slot_target_moved()));

	connect(lineEditCameraX, SIGNAL(returnPressed()), this, SLOT(slot_text_changed()));
	connect(lineEditCameraY, SIGNAL(returnPressed()), this, SLOT(slot_text_changed()));
	connect(lineEditCameraH, SIGNAL(returnPressed()), this, SLOT(slot_text_changed()));
	connect(lineEditTargetX, SIGNAL(returnPressed()), this, SLOT(slot_text_changed()));
	connect(lineEditTargetY, SIGNAL(returnPressed()), this, SLOT(slot_text_changed()));
	connect(lineEditTargetH, SIGNAL(returnPressed()), this, SLOT(slot_text_changed()));
}


/**
 * Window opening, move to saved position
 */
void main_window::showEvent(QShowEvent *)
{
	// Move the window to the appropriate place
	int x, y;
	my_settings->get_int2(WINDOW_POS_MAIN, &x, &y);
	if (x != -1 && y != -1) {
		move(x, y);
	}
}


/**
 * Restore windows which were open when program last exited
 */
void main_window::open_windows(void)
{
	bool open;

	my_settings->get_bool(WINDOW_RENDER, &open);
	if (open)	slot_render_win();
	my_settings->get_bool(WINDOW_PREVIEW, &open);
	if (open)	slot_preview_win();
	my_settings->get_bool(WINDOW_SETTINGS, &open);
	if (open)	slot_settings_win();
}


/**
 * Slots called by widgets
 */
void main_window::slot_perlin_iterations(int val)
{
	my_settings->set_int(PERLIN_ITERATIONS, int (val));
}


void main_window::slot_perlin_scale(int scale)
{
	my_settings->set_float(PERLIN_SCALE, scale);
}


void main_window::slot_perlin_persistence(int persistence)
{
	my_settings->set_float(PERLIN_PERSISTENCE, persistence);
}


void main_window::slot_perlin2_iterations(int val)
{
	my_settings->set_int(PERLIN2_OCTAVES, val);
}


void main_window::slot_perlin2_h(void)
{
	bool b;

	float val = w_perlin2->lineEditH->text().toFloat(&b);
	if (!b) return;

	my_settings->set_float(PERLIN2_FD, val);
}


void main_window::slot_perlin2_lacunarity(void)
{
	bool b;

	float val = w_perlin2->lineEditLacunarity->text().toFloat(&b);
	if (!b) return;

	my_settings->set_float(PERLIN2_LACUNARITY, val);
}


void main_window::slot_perlin2_scale(int scale)
{
	my_settings->set_float(PERLIN2_SCALE, scale);
}



void main_window::slot_multifractal_iterations(int val)
{
	my_settings->set_int(MULTIFRACTAL_ITERATIONS, val);
}


void main_window::slot_multifractal_h(void)
{
	bool b;

	float val = w_multifractal->lineEditH->text().toFloat(&b);
	if (!b) return;

	my_settings->set_float(MULTIFRACTAL_FD, val);
}


void main_window::slot_multifractal_lacunarity(void)
{
	bool b;

	float val = w_multifractal->lineEditLacunarity->text().toFloat(&b);
	if (!b) return;

	my_settings->set_float(MULTIFRACTAL_LACUNARITY, val);
}


void main_window::slot_multifractal_scale(int scale)
{
	my_settings->set_float(MULTIFRACTAL_SCALE, scale);
}


void main_window::slot_ridged_iterations(int val)
{
	my_settings->set_int(RIDGED_ITERATIONS, val);
}

void main_window::slot_ridged_lacunarity(void)
{
	bool b;

	float val = w_ridged->lineEditLacunarity->text().toFloat(&b);
	if (!b) return;

	my_settings->set_float(RIDGED_LACUNARITY, val);
}

void main_window::slot_ridged_h(void)
{
	bool b;

	float val = w_ridged->lineEditH->text().toFloat(&b);
	if (!b) return;

	my_settings->set_float(RIDGED_FD, val);
}

void main_window::slot_ridged_scale(int scale)
{
	my_settings->set_float(RIDGED_SCALE, scale);
}

void main_window::slot_ridged_gain(int gain)
{
	my_settings->set_float(RIDGED_GAIN, gain);
}



/**
 * Change fractal generation method
 */
void main_window::slot_hf_type(int i)
{
	switch (i) {
		case 0:		set_type(HF_PERLIN);				stack->raiseWidget(1);	break;
		case 1:		set_type(HF_PERLIN2);				stack->raiseWidget(2);	break;
		case 2:		set_type(HF_MULTIFRACTAL);			stack->raiseWidget(3);	break;
		case 3:		set_type(HF_RIDGED_PERLIN);			stack->raiseWidget(4);	break;
		default:
			printf("%d - Unknown fractal type!!", i);
			break;
	}
}


/**
 * Save rendered image to disk
 */
void main_window::slot_save_image()
{
	// Is there an image to save?
	bool val;
	my_settings->get_bool(IMAGE_TO_SAVE, &val);
	if (!val || !w_render) {
		KMessageBox::sorry(0, "There is no image to save.  Render one then try again.");
	} else {

		QString filename;
		filename = KFileDialog::getSaveFileName(
									QString::null,
									KImageIO::pattern(KImageIO::Writing),
									this);
		if (!filename.isNull()) {
			const QImage* image = w_render->get_bitmap();
			QPixmap* buffer = new QPixmap;
			buffer->convertFromImage(*image, 0);
			buffer->save(filename, KImageIO::type(filename).ascii());
			delete buffer;
		}
	}
}


/**
 * Calculate camera position in terrain space
 */
void main_window::calculate_camera_angle(void)
{
	float water_level;
	my_settings->get_float(WATER_LEVEL, &water_level);

	int camera_x, camera_z;

	// Get current camera positions
	relief_view->get_camera_pos(&camera_x, &camera_z);

	// Calculate heights of camera and target
	point3d p(camera_x, 0, camera_z);
	float height_camera = my_map->get_height(p.x, p.z);
	if (height_camera < 0) {
		height_camera = 0;
	}

	my_settings->set_float3(CAMERA_POSITION, camera_x, height_camera + 8, camera_z);
}


/**
 * Calculate target position in terrain space
 */
void main_window::calculate_target_angle(void)
{
	float water_level;
	my_settings->get_float(WATER_LEVEL, &water_level);

	int target_x, target_z;

	// Get current target position
	relief_view->get_target_pos(&target_x, &target_z);

	float height_target = my_map->get_height(target_x, target_z);
	if (height_target < 0) {
		height_target = 0;
	}

	my_settings->set_float3(TARGET_POSITION, target_x, height_target + 8, target_z);
}


/**
 * Update camera text display from relief_view
 */
void main_window::update_camera_text(void)
{
	int x, y;
	float height;

	relief_view->get_camera_pos(&x, &y);
	height = my_map->get_height(x, y);
	if (height < 0) height = 0;

	lineEditCameraX->setText(QString("%1").arg(float (x), 0, 'f', 1));
	lineEditCameraY->setText(QString("%1").arg(float (y), 0, 'f', 1));
	lineEditCameraH->setText(QString("%1").arg(float (height), 0, 'f', 1));
}


/**
 * Update target text display from relief_view
 */
void main_window::update_target_text(void)
{
	int x, y;
	float height;

	relief_view->get_target_pos(&x, &y);

	height = my_map->get_height(x, y);
	if (height < 0) height = 0;

	lineEditTargetX->setText(QString("%1").arg(float (x), 0, 'f', 1));
	lineEditTargetY->setText(QString("%1").arg(float (y), 0, 'f', 1));
	lineEditTargetH->setText(QString("%1").arg(float (height), 0, 'f', 1));
}


/**
 * Update camera coord display
 */
void main_window::update_text_display(void)
{
	update_target_text();
	update_camera_text();
}


/**
 * Camera moved, so inform any open windows
 */
void main_window::update_all(bool want_rerender, Update update)
{
	if (update == CAMERA || update == CAMERA_AND_TARGET) {
		calculate_camera_angle();
		update_camera_text();
	}

	if (update == TARGET || update == CAMERA_AND_TARGET) {
		calculate_target_angle();
		update_target_text();
	}

	my_settings->set_int(PREVIEW_SEED, get_seed());

	// Tell the other windows to update themselves..
	if (w_render) {
		w_render->new_map();
	}

	if (w_preview) {
		if (want_rerender) {
			w_preview->create_and_render();
		} else {
			w_preview->render();
		}
	}
}


/**
 * User closed window, so exit program
 */
void main_window::closeEvent(QCloseEvent *)
{
	slot_exit();
}


/**
 * Show 'About' message
 */
void main_window::slot_about_qt()
{
    QMessageBox::aboutQt( this, "Landscape" );
}



/**
 * Exit program
 */
void main_window::slot_exit()
{
	// Tell any open window to save its position
	if (w_render) {
		w_render->save_position();
	}
	if (w_preview) {
		w_preview->save_position();
	}
	if (w_settings) {
		w_settings->save_position();
	}
	save_position();

	kApp->quit();
}


/**
 * Create a new heightfield
 */
void main_window::slot_generate(void)
{
	my_settings->set_int(PREVIEW_HF_TYPE, get_type());
	set_seed(clock());
	generate_map();
}


/**
 * Open preview window
 */
void main_window::slot_preview_win()
{
	if (!w_preview) {
		w_preview = new preview_window(this);
		
		// Move the window to the appropriate place
		int x, y;
		my_settings->get_int2(WINDOW_POS_PREVIEW, &x, &y);
		if (x != -1 && y != -1) {
			w_preview->move(x, y);
		}
		w_preview->show();

		// Make sure we're notified when the window is closed
		connect(w_preview, SIGNAL(destroyed()), this, SLOT(slot_preview_closed()));

		w_preview->create_and_render();
		my_settings->set_bool(WINDOW_PREVIEW, true);
	} else {
		w_preview->show();
	}
}


/**
 * Open render window
 */
void main_window::slot_render_win()
{
	if (!w_render) {
		w_render = new render_window(this);

		// Move the window to the appropriate place
		int x, y;
		my_settings->get_int2(WINDOW_POS_RENDER, &x, &y);

		if (x != -1 && y != -1) {
			w_render->move(x, y);
		}

		w_render->show();

		// Make sure we here when the window is closed
		connect(w_render, SIGNAL(destroyed()), this, SLOT(slot_render_closed()));

		w_render->new_map();
		my_settings->set_bool(WINDOW_RENDER, true);
	} else {
		w_render->show();
	}
}


/**
 * Open user settings window
 */
void main_window::slot_settings_win()
{
	if (!w_settings) {

		w_settings = new settings_window(this);

		// Move the window to the appropriate place
		int x, y;
		my_settings->get_int2(WINDOW_POS_SETTINGS, &x, &y);
		if (x != -1 && y != -1) {
			w_settings->move(x, y);
		}
		w_settings->show();

		// Make sure we here when the window is closed
		connect(w_settings, SIGNAL(destroyed()), this, SLOT(slot_settings_closed()));

		my_settings->set_bool(WINDOW_SETTINGS, true);
	} else {
		w_settings->show();
	}
}


/**
 * Receive notification that render window has been closed
 */
void main_window::slot_render_closed()
{
	w_render = 0;
	my_settings->set_bool(WINDOW_RENDER, false);
}


/**
 * Receive notification that preview window has been closed
 */
void main_window::slot_preview_closed()
{
	w_preview = 0;
	my_settings->set_bool(WINDOW_PREVIEW, false);
}


/**
 * Receive notification that user settings window has been closed
 */
void main_window::slot_settings_closed()
{
	w_settings = 0;
	my_settings->set_bool(WINDOW_SETTINGS, false);
}


/**
 * Camera position has changed
 */
void main_window::slot_camera_moved(void)
{
	update_all(false, CAMERA);
}


/**
 * Target position has changed
 */
void main_window::slot_target_moved(void)
{
	update_all(false, TARGET);
}


/**
 * Generate height field and render to relief_view
 */
void main_window::generate_map(void)
{
	QPainter p(relief_view);

	// Get pointer to bitmap data
	QImage* bitmap = relief_view->get_bitmap_data();

	//Get pointer to current scanline
	uint32* p32 = (uint32*) bitmap->scanLine(0);
	int bpr = bitmap->bytesPerLine();

	render_relief(p32, bpr);

	// Draw the updated image
	relief_view->repaint(0, 0, RELIEF_SIZE, RELIEF_SIZE, false);

	update_all(true, CAMERA_AND_TARGET);
}


/**
 * User entered new data for camera or target, validate and update
 */
void main_window::slot_text_changed(void)
{
	// Reload to read any updated debug settings etc
	load_settings();

	float x, y, h;
	bool b;
	x = lineEditCameraX->text().toFloat(&b);
	if (!b) return;

	y = lineEditCameraY->text().toFloat(&b);
	if (!b) return;

	h = lineEditCameraH->text().toFloat(&b);
	if (!b) return;

	if (!unconstrained_camera) {
		if (x < -RELIEF_SIZE/2)		x = -RELIEF_SIZE/2;
		if (y < -RELIEF_SIZE/2)		y = -RELIEF_SIZE/2;
		if (x >  RELIEF_WIDTH/2)	x =  RELIEF_WIDTH/2;
		if (y >  RELIEF_WIDTH/2)	y =  RELIEF_WIDTH/2;

		lineEditCameraX->setText(QString("%1").arg(float (x), 0, 'f', 1));
		lineEditCameraY->setText(QString("%1").arg(float (y), 0, 'f', 1));
	}

	relief_view->set_camera_pos(int (x), int (y));
	my_settings->set_float3(CAMERA_POSITION, x, h + 8, y);


	x = lineEditTargetX->text().toFloat(&b);
	if (!b) return;

	y = lineEditTargetY->text().toFloat(&b);
	if (!b) return;

	h = lineEditTargetH->text().toFloat(&b);
	if (!b) return;

	if (!unconstrained_camera) {
		if (x < -RELIEF_SIZE/2)		x = -RELIEF_SIZE/2;
		if (y < -RELIEF_SIZE/2)		y = -RELIEF_SIZE/2;
		if (x >  RELIEF_WIDTH/2)	x =  RELIEF_WIDTH/2;
		if (y >  RELIEF_WIDTH/2)	y =  RELIEF_WIDTH/2;

		lineEditTargetX->setText(QString("%1").arg(float (x), 0, 'f', 1));
		lineEditTargetY->setText(QString("%1").arg(float (y), 0, 'f', 1));
	}

	relief_view->set_target_pos(int (x), int (y));
	my_settings->set_float3(TARGET_POSITION, x, h + 8, y);
}


/**
 * Update the fractal prefs widgets to reflect the current settings.
 */
void main_window::update_widgets(void)
{
	// Set widgets according to current settings
	int val;
	my_settings->get_int(PERLIN_ITERATIONS, &val);
	w_perlin->spinBoxIterations->setValue(val);

	my_settings->get_int(PERLIN2_OCTAVES, &val);
	w_perlin2->spinBoxOctaves->setValue(val);

	float f;
	QString str;

	my_settings->get_float(PERLIN2_FD, &f);
	str.setNum(f);
	w_perlin2->lineEditH->setText(str);
	my_settings->get_float(PERLIN2_LACUNARITY, &f);
	str.setNum(f);
	w_perlin2->lineEditLacunarity->setText(str);
	my_settings->get_float(PERLIN2_SCALE, &f);
	w_perlin2->sliderScale->setValue(f);

	my_settings->get_float(MULTIFRACTAL_FD, &f);
	str.setNum(f);
	w_multifractal->lineEditH->setText(str);
	my_settings->get_float(MULTIFRACTAL_LACUNARITY, &f);
	str.setNum(f);
	w_multifractal->lineEditLacunarity->setText(str);
	my_settings->get_float(MULTIFRACTAL_SCALE, &f);
	w_multifractal->sliderScale->setValue(f);
	my_settings->get_int(MULTIFRACTAL_ITERATIONS, &val);
	w_multifractal->spinBoxIterations->setValue(val);

	my_settings->get_float(RIDGED_FD, &f);
	str.setNum(f);
	w_ridged->lineEditH->setText(str);
	my_settings->get_float(RIDGED_LACUNARITY, &f);
	str.setNum(f);
	w_ridged->lineEditLacunarity->setText(str);
	my_settings->get_float(RIDGED_SCALE, &f);
	w_ridged->sliderScale->setValue(f);
	my_settings->get_float(RIDGED_GAIN, &f);
	w_ridged->sliderGain->setValue(f);
	my_settings->get_int(RIDGED_ITERATIONS, &val);
	w_ridged->spinBoxIterations->setValue(val);

	my_settings->get_float(PERLIN_SCALE, &f);
	w_perlin->sliderScale->setValue(f);
	my_settings->get_float(PERLIN_PERSISTENCE, &f);
	w_perlin->sliderPersistence->setValue(f);
}


/**
 * Store current window position
 */
void main_window::save_position(void)
{
	my_settings->set_int2(WINDOW_POS_MAIN, x(), y());
}
