//
// Qt widget which displays a representation of a camera view
//
// Copyright (C) J. Belson	2000.03.09
//
// $Author: jon $	:	$Date: 2003/06/08 15:50:43 $
//

#include <cstdio>
#include <cmath>
#include <qpainter.h>

#include "matrix.h"
#include "vector3d.h"
#include "mouse_bm_view.h"


/**
 * Constructor
 */
mouse_bm_view::mouse_bm_view(
					QWidget* parent,
					const char* name,
					int width,
					int height)
	: QWidget(parent, name),
	  WIDTH(width),
	  tracking(TRACKING_NONE)
{
	image = new QImage(width, height, 32);

	reset_camera();
}


/**
 * Destructor
 */
mouse_bm_view::~mouse_bm_view()
{
	delete image;
}


/**
 * Draw line from camera to target
 */
void mouse_bm_view::draw_line_of_sight(void)
{
	QPainter p(this);
	p.setPen(SolidLine);
	p.setBrush(QColor(0, 0, 0));
	p.setRasterOp(NotXorROP /*NoPen*/);

	const float fov = 0.4;

	// Draw left/right bounds of line of sight...
	vector3d v2(target_x - camera_x, 0.0, target_y - camera_y);

	vector3d lh = matrix(matrix::ROTATION, 0, -fov, 0)*v2;
	p.drawLine(camera_x, camera_y,  camera_x + int(lh.i), camera_y + int(lh.k));

	vector3d rh = matrix(matrix::ROTATION, 0, fov, 0)*v2;
	p.drawLine(camera_x, camera_y,  camera_x + int(rh.i), camera_y + int(rh.k));


	// Draw a cross at the target...
	matrix m = matrix(matrix::ROTATION, 0, M_PI/2.0, 0);

	vector3d v(target_x - camera_x, 0.0, target_y - camera_y, true);
	v = v*10;
	v = matrix(matrix::ROTATION, 0, M_PI/4.0, 0)*v;

	for (int i=0; i<4; i++) {
		p.drawLine(target_x, target_y,  target_x + int(v.i), target_y + int(v.k));
		v = m*v;
	}


	p.setRasterOp(CopyROP);

}


/**
 * Move camera to default position and orientation
 */
void mouse_bm_view::reset_camera(void)
{
	camera_x = WIDTH/2;
	camera_y = WIDTH - 1;

	target_x = WIDTH/2;
	target_y = 0;

#if 0
	camera_x = 0;
	camera_y = -WIDTH/2;
	target_x = 0;
	target_y = 0;
#endif
}


/**
 * Return current camera position, adjusted so (0, 0) is at the
 *  centre.
 */
void mouse_bm_view::get_camera_pos(int *x, int *y)
{
	*x = camera_x - WIDTH/2;
	*y = camera_y - WIDTH/2;
}


/**
 * Position camera at new location
 */
void mouse_bm_view::set_camera_pos(int x, int y)
{
	draw_line_of_sight();

	camera_x = x + WIDTH/2;
	camera_y = y + WIDTH/2;

	draw_line_of_sight();
}


/**
 * Return current target position, adjusted so (0, 0) is at the
 *  centre.
 */
void mouse_bm_view::get_target_pos(int *x, int *y)
{
	*x = target_x - WIDTH/2;
	*y = target_y - WIDTH/2;
}


/**
 * Point camera at new location
 */
void mouse_bm_view::set_target_pos(int x, int y)
{
	draw_line_of_sight();

	target_x = x + WIDTH/2;
	target_y = y + WIDTH/2;

	draw_line_of_sight();
}


/**
 * Camera or target has moved, so update and inform parent window
 */
void mouse_bm_view::update(int x, int y)
{
	switch (tracking) {
		case TRACKING_CAMERA:
			draw_line_of_sight();
			camera_x = x;
			camera_y = y;
			draw_line_of_sight();
			emit signal_camera_moved();
			break;

		case TRACKING_TARGET:
			draw_line_of_sight();
			target_x = x;
			target_y = y;
			draw_line_of_sight();
			emit signal_target_moved();
			break;

		default:
			;
	}

}


/**
 * Track mouse movement
 */
void mouse_bm_view::mousePressEvent(QMouseEvent* e)
{
	if (e->button() == LeftButton) {
		tracking = TRACKING_CAMERA;
		update(e->x(), e->y());
	}
	else if (e->button() == RightButton) {
		tracking = TRACKING_TARGET;
		update(e->x(), e->y());
	}
}


/**
 * User release mouse over widget
 */
void mouse_bm_view::mouseReleaseEvent(QMouseEvent* e)
{
	int x = e->x();
	int y = e->y();

	if (x<0) {
		x = 0;
	}
	if (x>WIDTH) {
		x = WIDTH-1;
	}
	if (y<0) {
		y = 0;
	}
	if (y>WIDTH) {
		y = WIDTH-1;
	}
	update(x, y);
	tracking = TRACKING_NONE;
}


/**
 * User pressed mouse over widget
 */
void mouse_bm_view::mouseMoveEvent(QMouseEvent* e)
{
	int x = e->x();
	int y = e->y();

	if (x<0) {
		x = 0;
	}
	if (x>WIDTH) {
		x = WIDTH-1;
	}
	if (y<0) {
		y = 0;
	}
	if (y>WIDTH) {
		y = WIDTH-1;
	}


//	if (x<0 || x>WIDTH || y<0 || y>WIDTH) {
//		return;
//	}

	switch (tracking) {
		case TRACKING_CAMERA:
		case TRACKING_TARGET:
			update(x, y);
			break;

		default:
			;
	}
}


/**
 * Redraw the screen
 */
void mouse_bm_view::paintEvent(QPaintEvent* e)
{
	QPainter p(this);
	//p.drawImage(0, 0, *image);

	QRect r = e->rect();
	p.drawImage(r.x(), r.y(), *image, r.x(), r.y(), r.width(), r.height());

	draw_line_of_sight();
}



