//
// render_window.cc
//
// Copyright (C) J. Belson	2000.10.08
//
// $Author: jon $	: $Date: 2003/05/30 11:04:04 $
//

#include <sys/types.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>

#include <qcheckbox.h>
#include <qcombobox.h>
#include <qpushbutton.h>
#include <qslider.h>


#include "bm_window.h"
#include "progress.h"
#include "renderer_flat_int.h"
#include "render_window.h"
#include "scene.h"
#include "settings.h"


using std::cout;
using std::endl;


static struct {
	int width, height;
} w_size[] = {
	{ 320,  240 },
	{ 640,  480 },
	{ 800,  600 },
	{ 1024, 768 }
};


/**
 * Constructor
 */
render_window::render_window(QWidget* )
	: widget_render(),
	  base_render_window(),
	  rendering(false)
{
	setCaption("Render window");

	w_size_index = 1;

	num_threads = 1;

	//cb_reflections->setDisabled(true);
	checkBoxTrees->setDisabled(true);

	comboBoxThreads->setDisabled(true);
	//combo_update->setDisabled(true);

	comboBoxSize->setCurrentItem(w_size_index);

	connect(comboBoxSize, SIGNAL(activated(int)), this, SLOT(slot_size(int)));
	connect(comboBoxUpdate, SIGNAL(activated(int)), this, SLOT(slot_update(int)));
	connect(pushButtonRender, SIGNAL(clicked()), this, SLOT(slot_render()));
	connect(sliderDetail, SIGNAL(valueChanged(int)), this, SLOT(slot_detail(int)));
	connect(checkBoxBackfaceCull, SIGNAL(clicked()), this, SLOT(slot_backface_cull()));
	connect(checkBoxFrustrumCull, SIGNAL(clicked()), this, SLOT(slot_frustrum_cull()));

	connect(checkBoxLighting, SIGNAL(clicked()), this, SLOT(slot_lighting()));
	connect(checkBoxAtmospherics, SIGNAL(clicked()), this, SLOT(slot_atmospherics()));
	connect(checkBoxTextures, SIGNAL(clicked()), this, SLOT(slot_textures()));
	connect(checkBoxShadows, SIGNAL(clicked()), this, SLOT(slot_shadows()));
	connect(checkBoxReflections, SIGNAL(clicked()), this, SLOT(slot_reflections()));
	connect(checkBoxSky, SIGNAL(clicked()), this, SLOT(slot_sky()));
	connect(checkBoxGround, SIGNAL(clicked()), this, SLOT(slot_ground()));

	int val;
	my_settings->get_int(RENDER_DETAIL, &val);
	sliderDetail->setValue(val);

	my_settings->get_bool(RENDER_BACKFACE_CULL, &backface_cull);
	checkBoxBackfaceCull->setChecked(backface_cull);

	my_settings->get_bool(RENDER_FRUSTRUM_CULL, &frustrum_cull);
	checkBoxFrustrumCull->setChecked(frustrum_cull);

	checkBoxLighting->setChecked(lighting);
	checkBoxSky->setChecked(sky);
	checkBoxGround->setChecked(ground);
	checkBoxAtmospherics->setChecked(atmospherics);
	checkBoxTextures->setChecked(textures);
	checkBoxReflections->setChecked(reflections);
	checkBoxTrees->setChecked(trees);
	checkBoxShadows->setChecked(shadows);
	comboBoxUpdate->setCurrentItem(3);

	open_output_window();

}



/**
 * Create display window, using w_width/w_height
 */
void render_window::open_output_window(void)
{
	int w_width = w_size[w_size_index].width;
	int w_height = w_size[w_size_index].height;

	// Open output window...
	win = new bm_window(w_width, w_height);

	// Move the window to the appropriate place
	int x, y;
	my_settings->get_int2(WINDOW_POS_RENDER_OUTPUT, &x, &y);
	if (x != -1 && y != -1) {
		win->move(x, y);
	}

	win->set_size(w_width, w_height);
	win->setCaption("Render Output");
	win->show();
}


/**
 * A new map was generated, get the new seed
 */
void render_window::new_map(void)
{
	cout << "render_window got new_map" << endl;		

	my_settings->get_int(PREVIEW_SEED, &seed);
}


/**
 * Launch the rendering code.  Used threads under BeOS, but here we
 *  just use a function call.
 */
void render_window::xrun(void)
{
	if (rendering) {
		cout << "Warning  - tried to render, but we're rendering already!!" << endl;
	} else {
		rendering = true;

		//launch_thread(this); 

		main_render_thread();

		//rendering = false;
	}
}


/**
 * Call render thread via 'this' pointer
 * @param arg pointer to render_window object
 */
void render_window::launch_thread(void* arg)
{
	static_cast<render_window*>(arg)->main_render_thread();
}


/**
 * Render the image according to user's settings
 */
void render_window::main_render_thread(void)
{
	// Stop user from changing render settings while we're rendering
	disable_widgets();

	// Get requested number of threads (can't change during a render ;^)
	//int active_threads = num_threads;

	// Start redraws
	win->set_redraw_rate(redraw_rate);

	win->clear_window();

	// Fast updates when rendering...
	win->set_smooth_update(false);

	scene* my_scene = new scene(win);
	int val;
	my_settings->get_int(RENDER_DETAIL, &val);
	my_scene->set_detail(val);

	float water_level;
	my_settings->get_float(WATER_LEVEL, &water_level);

	if (ground) {
		int type;
		my_settings->get_int(PREVIEW_HF_TYPE, &type);
		my_map = new hfield(hf_type(type));
		my_map->set_offset(-water_level);		// Set zero point to water level
		my_scene->set_land(my_map);
		my_map->seed_rand(seed);
	}

	if (sky) {
		my_scene->set_sky(true);
	}

	my_scene->set_atmospherics(atmospherics);
	my_scene->set_textures(textures);
	my_scene->set_shadows(shadows);
	my_scene->set_reflections(reflections);

	// Get camera position
	float cam_x, cam_y, cam_z;
	float tar_x, tar_y, tar_z;

	my_settings->get_float3(CAMERA_POSITION, &cam_x, &cam_y, &cam_z);
	my_settings->get_float3(TARGET_POSITION, &tar_x, &tar_y, &tar_z);

	// Make vector in view direction
	vector3d heading(tar_x - cam_x, tar_y - cam_y, tar_z - cam_z);
	heading.normalise();

	my_scene->set_view(point3d(cam_x, cam_y, cam_z), heading);

	my_scene->render();

	delete my_scene;


	rendering = false;
	puts("Finished render");

	// We appear to have rendered an image, so...
	my_settings->set_bool(IMAGE_TO_SAVE, true);

	// Disable automatic window update
	win->set_redraw_rate(0);

	win->set_smooth_update(true);

	//	win->repaint(0, 0, w_width, w_height, false);
	win->slot_refresh();
	
	// Reenable render widgets
	enable_widgets();
}


/**
 * Render option widgets should be disabled while rendering
 */
void render_window::disable_widgets(void)
{
	pushButtonRender->setEnabled(false);
	comboBoxSize->setEnabled(false);
	sliderDetail->setEnabled(false);
}


/**
 * Rendering option widgets should be disabled while rendering
 */
void render_window::enable_widgets(void)
{
	pushButtonRender->setEnabled(true);
	comboBoxSize->setEnabled(true);
	sliderDetail->setEnabled(true);
}


/**
 * Window close event
 */
void render_window::closeEvent(QCloseEvent *)
{
	// Save window position
	save_position();

	delete win;
	delete this;
}


/**
 * Hide output window when main render window is iconified
 */
void render_window::hideEvent(QHideEvent* )
{
	win->hide();
}


/**
 * Show output window when main window is shown
 */
void render_window::showEvent(QShowEvent* )
{
	win->show();
}


/**
 * Window close slot
 */
void render_window::slot_close()
{
	puts("Close me");
}


/**
 * Window exuit slot
 */
void render_window::slot_exit()
{
	close();
}


/**
 * Redner button slot
 */
void render_window::slot_render(void)
{
	puts("Render");

	if (rendering) {
		cout << "Warning  - tried to render, but we're rendering already!!" << endl;
	} else {
		rendering = true;
		main_render_thread();
	}
}


/**
 * Render output size changed
 */
void render_window::slot_size(int current)
{
	if (current != w_size_index) {
		w_width	= w_size[current].width;
		w_height	= w_size[current].height;

		win->set_size(w_width, w_height);	//setFixedSize(width, height);
		
		w_size_index = current;
	}
}


/**
 * Render detail slider changed
 */
void render_window::slot_detail(int num)
{
	cout << "Detail = " << num << endl;
	detail = num;
	my_settings->set_int(RENDER_DETAIL, num);
}


/**
 * Backface culling checkbox clicked
 */
void render_window::slot_backface_cull(void)
{
	backface_cull = !backface_cull;
	my_settings->set_bool(RENDER_BACKFACE_CULL, backface_cull);
}


/**
 * Frustrum culling checkbox clicked
 */
void render_window::slot_frustrum_cull(void)
{
	frustrum_cull = !frustrum_cull;
	my_settings->set_bool(RENDER_FRUSTRUM_CULL, frustrum_cull);
}


/**
 * Update rate changed
 */
void render_window::slot_update(int num)
{
	switch (num) {
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
			redraw_rate = num;
			break;
		default:
			puts("Unknown update rate!");
	}
}


/**
 * Lighting checkbox clicked
 */
void render_window::slot_lighting()
{
	puts("lighting");
	lighting = !lighting;
	my_settings->set_bool(LIGHTING, lighting);
}


/**
 * Toggle sky rendering
 */
void render_window::slot_sky()
{
	sky = !sky;
}


/**
 * Toggle ground rendering
 */
void render_window::slot_ground()
{
	ground = !ground;
}


/**
 * Atmospherics checkbox clicked
 */
void render_window::slot_atmospherics()
{
	atmospherics = !atmospherics;
	my_settings->set_bool(ATMOSPHERICS, atmospherics);
}


/**
 * Texture checkbox clicked
 */
void render_window::slot_textures()
{
	textures = !textures;
}


/**
 * Reflection checkbox clicked
 */
void render_window::slot_reflections()
{
	reflections = !reflections;
}


/**
 * Tree checkbox clicked
 */
void render_window::slot_trees()
{
	;
}


/**
 * Shadow checkbox clicked
 */
void render_window::slot_shadows()
{
	shadows = !shadows;
	checkBoxShadows->setChecked(shadows);
}


/**
 * Get pointer to rendering image data
 */
const QImage* render_window::get_bitmap(void)
{
	return win->get_bitmap();
}


/**
 * Move render control window
 * @param x x coordinate
 * @param y y coordinate
 */
void render_window::set_position(int x, int y)
{
	move(x, y);
}


/**
 * Store position of render contral and output windows
 */
void render_window::save_position(void)
{
	puts("render_window::save_position()");

	my_settings->set_int2(WINDOW_POS_RENDER, x(), y());
	my_settings->set_int2(WINDOW_POS_RENDER_OUTPUT, win->x(), win->y());
}
